"""
Planar Point Pattern Class

"""
import numpy as np
import sys
from pysal.lib.cg import KDTree
from .centrography import hull
from .window import as_window,  poly_from_bbox
from .util import cached_property
import pandas as pd
from matplotlib import pyplot as plt
from matplotlib.collections import PatchCollection
from matplotlib.patches import Polygon

__author__ = "Serge Rey sjsrey@gmail.com"
__all__ = ['PointPattern']

if sys.version_info[0] > 2:
    xrange = range


class PointPattern(object):
    """
    Planar Point Pattern Class 2-D.

    Parameters
    ----------
    points:       array
                  (n,p), n points with p >= 2 attributes on each
                  point. Two attributes must comprise the spatial
                  coordinate pair. Default is that the first two
                  attributes are the x and y spatial coordinates.
    window:       :class:`.Window`
                  Bounding geometric object for the point pattern.
                  If not specified, window will be set to the minimum
                  bounding rectangle of the point pattern.
    names:        list
                  The names of the attributes.
    coord_names:  list
                  The names of the attributes defining the two spatial
                  coordinates.

    Examples
    --------
    >>> from pointpats import PointPattern
    >>> points = [[66.22, 32.54], [22.52, 22.39], [31.01, 81.21],
    ...           [9.47, 31.02], [30.78, 60.10], [75.21, 58.93],
    ...           [79.26,  7.68], [8.23, 39.93], [98.73, 77.17],
    ...           [89.78, 42.53], [65.19, 92.08], [54.46, 8.48]]
    >>> pp = PointPattern(points)
    >>> pp.n
    12
    >>> pp.mean_nnd
    21.612139802089246
    >>> pp.lambda_mbb
    0.0015710507711240867
    >>> pp.lambda_hull
    0.0022667153468973137
    >>> pp.hull_area
    5294.00395
    >>> pp.mbb_area
    7638.200000000001

    """
    def __init__(self, points, window=None, names=None, coord_names=None):

        # first two series in df are x, y unless coor_names and names are
        # specified

        self.df = pd.DataFrame(points)
        n, p = self.df.shape
        self._n_marks = p - 2
        if coord_names is None:
            if names is not None:
                coord_names = names[:2]
            else:
                coord_names = ['x', 'y']
        if names is None:
            col_names = coord_names
            if p > 2:
                for m in range(2, p):
                    col_names.append("mark_{}".format(m-2))
            coord_names = coord_names[:2]
        else:
            col_names = names

        self.coord_names = coord_names
        self._x, self._y = coord_names
        self.df.columns = col_names
        self.points = self.df.loc[:, [self._x, self._y]]
        self._n, self._p = self.points.shape
        if window is None:
            self.set_window(as_window(poly_from_bbox(self.mbb)))
        else:
            self.set_window(window)

        self._facade()

    def __len__(self):
        """Return the number of points. Use the expression 'len(pp)'.

        Returns
        -------
        length : int
            The number of points in the point pattern.

        Examples
        --------
        >>> from pointpats import PointPattern
        >>> points = [[1, 3], [4, 5], [0,0]]
        >>> pp = PointPattern(points)
        >>> len(pp)
        3

        """
        return len(self.df)

    def __contains__(self, n):
        """Return True if n is a point (a tuple of coordinates), False otherwise.
        Use the expression 'n in pp'.

        Examples
        --------
        >>> from pointpats import PointPattern
        >>> points = [[1, 3], [4, 5], [0,0]]
        >>> pp = PointPattern(points)
        >>> [1, 3] in pp
        True
        """
        name = self.df.columns.values.tolist()
        return ((self.df[name[0]] == n[0]) & (self.df[name[1]] == n[1])).any()

    def set_window(self, window):
        try:
            self._window = window
        except:
            print("not a valid Window object")

    def get_window(self):
        """
        Bounding geometry for the point pattern

        :class:`.window.Window`
        """
        if not hasattr(self, '_window') or self._window is None:
            # use bbox as window
            self.set_window(as_window(poly_from_bbox(self.mbb)))
        return self._window

    window = property(get_window, set_window)

    def summary(self):
        '''
        Description of the point pattern.
        '''

        print('Point Pattern')
        print("{} points".format(self.n))
        print("Bounding rectangle [({},{}), ({},{})]".format(*self.mbb))
        print("Area of window: {}".format(self.window.area))
        print("Intensity estimate for window: {}".format(self.lambda_window))
        print(self.head())

    def add_marks(self, marks, mark_names=None):
        if mark_names is None:
            nm = range(len(marks))
            mark_names = ["mark_{}".format(self._n_marks+1+j) for j in nm]
        for name, mark in zip(mark_names, marks):
            self.df[name] = mark
            self._n_marks += 1

    def plot(self, window=False, title="Point Pattern", hull=False,
             get_ax=False):
        """
        Plot function for a point pattern.

        Parameters
        ----------
        window  : boolean
                  If window is True, plot window of the point
                  pattern. If not, don't plot window.
        title   : string
                  Name of the figure.
        hull    : boolean
                  If hull is True, plot convex hull of the point
                  pattern. If not, don't plot convex hull.
        get_ax  : boolean
                  If get_ax is True, return the current plot ax.
                  
        Returns
        -------
        ax      : matplotlib.axes._subplots.AxesSubplot
                  Current plot ax. Only return it when get_ax is True.

        """
        fig, ax = plt.subplots()
        plt.plot(self.df[self._x], self.df[self._y], '.')
        # plt.scatter(self.df[self._x], self.df[self._y])
        plt.title(title)
        if window:
            patches = []
            for part in self.window.parts:
                p = Polygon(np.asarray(part))
                patches.append(p)
            ax.add_collection(PatchCollection(patches, facecolor='w',
                              edgecolor='k', alpha=0.3))
        if hull:
            patches = []
            p = Polygon(self.hull)
            patches.append(p)
            ax.add_collection(PatchCollection(patches, facecolor='w',
                              edgecolor='k', alpha=0.3))

        # plt.plot(x, y, '.')
        if get_ax:
            return ax

    def _mbb(self):
        """
        Minimum bounding box
        """
        mins = self.points.min(axis=0)
        maxs = self.points.max(axis=0)
        return np.hstack((mins, maxs))

    mbb = cached_property(_mbb)

    def _mbb_area(self):
        """
        Area of minimum bounding box
        """

        return np.product(self.mbb[[2, 3]]-self.mbb[[0, 1]])

    mbb_area = cached_property(_mbb_area)

    def _n(self):
        """
        Number of points
        """
        return self.points.shape[0]

    n = cached_property(_n)

    def _lambda_mbb(self):
        """
        Intensity based on minimum bounding box
        """
        return self.n * 1. / self.mbb_area

    lambda_mbb = cached_property(_lambda_mbb)

    def _hull(self):
        """
        Points defining convex hull in counterclockwise order
        """
        return hull(self.points)

    hull = cached_property(_hull)

    def _lambda_window(self):
        """
        Intensity estimate based on area of window

        The intensity of a point process at point :math:`s_j` can be defined
        as:

        .. math::

            \\lambda(s_j) = \\lim \\limits_{|\\mathbf{A}s_j|
            \\to 0} \\left \\{ \\frac{E(Y(\mathbf{A}s_j)}{|\mathbf{A}s_j|}
            \\right \\}

        where :math:`\\mathbf{A}s_j` is a small region surrounding location
        :math:`s_j` with area :math:`|\\mathbf{A}s_j|`, and
        :math:`E(Y(\\mathbf{A}s_j))` is the expected number of event points in
        :math:`\\mathbf{A}s_j`.

        The intensity is the mean number of event points per unit of area at
        point :math:`s_j`.

        """
        return self.n / self.window.area

    lambda_window = cached_property(_lambda_window)

    def _hull_area(self):
        """
        Area of convex hull
        """
        h = self.hull
        if not np.alltrue(h[0] == h[-1]):
            # not in closed cartographic form
            h = np.vstack((h, h[0]))
        s = h[:-1, 0] * h[1:, 1] - h[1:, 0] * h[:-1, 1]
        return s.sum() / 2.

    hull_area = cached_property(_hull_area)

    def _lambda_hull(self):
        """
        Intensity based on convex hull
        """
        return self.n * 1. / self.hull_area

    lambda_hull = cached_property(_lambda_hull)

    def _build_tree(self):
        return KDTree(self.points)
    tree = cached_property(_build_tree)

    def knn(self, k=1):
        """
        Find k nearest neighbors for each point in the pattern

        Parameters
        ----------
        k:      int
                number of nearest neighbors to find

        Returns
        -------
        nn:    array (n x k)
               row i  column j contains the id for i's jth nearest neighbor

        nnd:   array(n x k)
               row i column j contains the distance between i and its jth
               nearest neighbor
        """
        if k < 1:
            raise ValueError('k must be at least 1')
        nn = self.tree.query(self.tree.data, k=k+1)
        return nn[1][:, 1:], nn[0][:, 1:]

    def _nn_sum(self):
        """
        Nearest neighbor distances
        """
        ids, nnd = self.knn(1)
        return nnd

    nnd = cached_property(_nn_sum)  # nearest neighbor distances

    def _min_nnd(self):
        """
        Min nearest neighbor distance
        """
        return self.nnd.min()

    min_nnd = cached_property(_min_nnd)

    def _max_nnd(self):
        """
        Max nearest neighbor distance
        """
        return self.nnd.max()

    max_nnd = cached_property(_max_nnd)

    def _mean_nnd(self):
        """
        Mean nearest neighbor distance
        """
        return self.nnd.mean()

    mean_nnd = cached_property(_mean_nnd)

    def find_pairs(self, r):
        """
        Find all pairs of points in the pattern that are within r units of each
        other

        Parameters
        ----------
        r: float
           diameter of pair circle

        Returns
        -------
        s: set
           pairs of points within r units of each other

        """
        return self.tree.query_pairs(r)

    def knn_other(self, other, k=1):
        """
        Find k nearest neighbors in the pattern for each point in other

        Parameters
        ----------
        other: :class:`PointPattern`

        k:      int
                number of nearest neighbors to find

        Returns
        -------
        nn:    array (n x k)
               row i  column j contains the id for i's jth nearest neighbor

        nnd:   array(n x k)
               row i column j contains the distance between i and its jth
               nearest neighbor
        """
        if k < 1:
            raise ValueError('k must be at least 1')
        try:
            nn = self.tree.query(other.points, k=k)
        except:
            nn = self.tree.query(other, k=k)
        return nn[1], nn[0]

    def explode(self, mark):
        """
        Explode a marked point pattern into a sequence of individual point
        patterns. If the mark has k unique values, then the sequence will be of
        length k.

        Parameters
        ----------
        mark: string
              The label of the mark to use for the subsetting

        Returns
        -------
        pps: list
             sequence of :class:`PointPattern` instances
        """

        uv = np.unique(self.df[mark])
        pps = [self.df[self.df[mark] == v] for v in uv]
        names = self.df.columns.values.tolist()
        cnames = self.coord_names
        return[PointPattern(pp, names=names, coord_names=cnames) for pp in pps]

    def unique(self):
        """ Remove duplicate points in the point pattern.

        Two points in a point pattern are deemed to be identical if their
        coordinates are the same, and their marks are the same (if any)

        Returns
        -------
        pp: list
            A deduplicated :class:`PointPattern` instance

        Examples
        --------
        >>> from pointpats import PointPattern
        >>> points = [[1.2, 2.1], [1.2, 2.1], [0, 1], [1, 2]]
        >>> pp = PointPattern(points)
        >>> pp.unique().df
             x    y
        0  1.2  2.1
        2  0.0  1.0
        3  1.0  2.0
        """
        names = self.df.columns.values.tolist()
        coord_names = self.coord_names
        window = self.set_window
        unique_df = self.df.drop_duplicates()
        return PointPattern(unique_df, names=names, coord_names=coord_names,
                            window=window)

    def superimpose(self, point_pattern):
        """Returns a superimposed point pattern.

        Parameters
        ----------
        point_pattern:
              :class:`PointPattern` instance

        Returns
        -------
        superimposed :
            :class:`PointPattern` instance

        Examples
        --------
        >>> from pointpats import PointPattern
        >>> points1 = [[1, 3], [4, 5], [0, 0]]
        >>> points2 = [[5, 6], [1, 4], [0, 0]]
        >>> pp1 = PointPattern(points1)
        >>> pp2 = PointPattern(points2)
        >>> pp1.superimpose(pp2).points
           x  y
        0  1  3
        1  4  5
        2  0  0
        0  5  6
        1  1  4
        """
        names_pp1 = self.df.columns.values.tolist()
        cnames_pp1 = self.coord_names
        names_pp2 = point_pattern.df.columns.values.tolist()
        cnames_pp2 = point_pattern.coord_names
        if names_pp1 != names_pp2 or cnames_pp1 != cnames_pp2:
            raise TypeError('Both point patterns should have similar\
                            attributes and spatial coordinates ')
        pp = pd.concat((self.df, point_pattern.df))
        pp = pp.drop_duplicates()
        return PointPattern(pp, names=names_pp1, coord_names=cnames_pp1)

    def flip_coordinates(self):
        """ Flips the coordinates of a point pattern.

        Doesn't change the structure of data frame. This function swaps
        `_x` and `_y` variables, which are used to represent coordinates.
        """
        self._x, self._y = self._y, self._x

    # Pandas facade
    def _facade(self):
            self.head = self.df.head
            self.tail = self.df.tail



"""
Centrographic measures for point patterns

TODO

- testing
- documentation

"""

__author__ = "Serge Rey sjsrey@gmail.com"
__all__ = ['mbr', 'hull', 'mean_center', 'weighted_mean_center',
           'manhattan_median', 'std_distance', 'euclidean_median', 'ellipse',
           'skyum', 'dtot',"_circle"]


import sys
import numpy as np
import warnings
import copy
from math import pi as PI
from scipy.spatial import ConvexHull
from pysal.lib.cg import get_angle_between, Ray, is_clockwise
from scipy.spatial import distance as dist
from scipy.optimize import minimize

not_clockwise = lambda x: not is_clockwise(x)

MAXD = sys.float_info.max
MIND = sys.float_info.min


def mbr(points):
    """
    Find minimum bounding rectangle of a point array.

    Parameters
    ----------
    points : arraylike
             (n,2), (x,y) coordinates of a series of event points.

    Returns
    -------
    min_x  : float
             leftmost value of the vertices of minimum bounding rectangle.
    min_y  : float
             downmost value of the vertices of minimum bounding rectangle.
    max_x  : float
             rightmost value of the vertices of minimum bounding rectangle.
    max_y  : float
             upmost value of the vertices of minimum bounding rectangle.

    """
    points = np.asarray(points)
    min_x = min_y = MAXD
    max_x = max_y = MIND
    for point in points:
        x, y = point
        if x > max_x:
            max_x = x
        if x < min_x:
            min_x = x
        if y > max_y:
            max_y = y
        if y < min_y:
            min_y = y
    return min_x, min_y, max_x, max_y


def hull(points):
    """
    Find convex hull of a point array.

    Parameters
    ----------
    points: arraylike
            (n,2), (x,y) coordinates of a series of event points.

    Returns
    -------
    _     : array
            (h,2), points defining the hull in counterclockwise order.
    """

    points = np.asarray(points)
    h = ConvexHull(points)
    return points[h.vertices]


def mean_center(points):
    """
    Find mean center of a point array.

    Parameters
    ----------
    points: arraylike
            (n,2), (x,y) coordinates of a series of event points.

    Returns
    -------
    _     : array
            (2,), (x,y) coordinates of the mean center.
    """

    points = np.asarray(points)
    return points.mean(axis=0)


def weighted_mean_center(points, weights):
    """
    Find weighted mean center of a marked point pattern.

    Parameters
    ----------
    points  : arraylike
              (n,2), (x,y) coordinates of a series of event points.
    weights : arraylike
              a series of attribute values of length n.

    Returns
    -------
    _      : array
             (2,), (x,y) coordinates of the weighted mean center.
    """


    points, weights = np.asarray(points), np.asarray(weights)
    w = weights * 1. / weights.sum()
    w.shape = (1, len(points))
    return np.dot(w, points)[0]


def manhattan_median(points):
    """
    Find manhattan median of a point array.

    Parameters
    ----------
    points  : arraylike
              (n,2), (x,y) coordinates of a series of event points.

    Returns
    -------
    _      : array
             (2,), (x,y) coordinates of the manhattan median.
    """

    points = np.asarray(points)
    if not len(points) % 2:
        s = "Manhattan Median is not unique for even point patterns."
        warnings.warn(s)
    return np.median(points, axis=0)


def std_distance(points):
    """
    Calculate standard distance of a point array.

    Parameters
    ----------
    points  : arraylike
              (n,2), (x,y) coordinates of a series of event points.

    Returns
    -------
    _      : float
             standard distance.
    """

    points = np.asarray(points)
    n, p = points.shape
    m = points.mean(axis=0)
    return np.sqrt(((points*points).sum(axis=0)/n - m*m).sum())


def ellipse(points):
    """
    Calculate parameters of standard deviational ellipse for a point pattern.

    Parameters
    ----------
    points : arraylike
             (n,2), (x,y) coordinates of a series of event points.

    Returns
    -------
    _      : float
             semi-major axis.
    _      : float
             semi-minor axis.
    theta  : float
             clockwise rotation angle of the ellipse.

    Notes
    -----
    Implements approach from:

    https://www.icpsr.umich.edu/CrimeStat/files/CrimeStatChapter.4.pdf
    """

    points = np.asarray(points)
    n, k = points.shape
    x = points[:, 0]
    y = points[:, 1]
    xd = x - x.mean()
    yd = y - y.mean()
    xss = (xd * xd).sum()
    yss = (yd * yd).sum()
    cv = (xd * yd).sum()
    num = (xss - yss) + np.sqrt((xss - yss)**2 + 4 * (cv)**2)
    den = 2 * cv
    theta = np.arctan(num / den)
    cos_theta = np.cos(theta)
    sin_theta = np.sin(theta)
    n_2 = n - 2
    sd_x = (2 * (xd * cos_theta - yd * sin_theta)**2).sum() / n_2
    sd_y = (2 * (xd * sin_theta - yd * cos_theta)**2).sum() / n_2
    return np.sqrt(sd_x), np.sqrt(sd_y), theta


def dtot(coord, points):
    """
    Sum of Euclidean distances between event points and a selected point.

    Parameters
    ----------
    coord   : arraylike
              (x,y) coordinates of a point.
    points  : arraylike
              (n,2), (x,y) coordinates of a series of event points.

    Returns
    -------
    d       : float
              sum of Euclidean distances.

    """
    points = np.asarray(points)
    xd = points[:, 0] - coord[0]
    yd = points[:, 1] - coord[1]
    d = np.sqrt(xd*xd + yd*yd).sum()
    return d

def euclidean_median(points):
    """
    Calculate the Euclidean median for a point pattern.

    Parameters
    ----------
    points: arraylike
            (n,2), (x,y) coordinates of a series of event points.

    Returns
    -------
    _     : array
            (2,), (x,y) coordinates of the Euclidean median.

    """
    points = np.asarray(points)
    start = mean_center(points)
    res = minimize(dtot, start, args=(points,))
    return res['x']

def skyum(points, not_hull=True):
    """
    Implements Skyum (1990)'s algorithm for the minimum bounding circle in R^2.

    0. Store points clockwise.
    1. Find p in S that maximizes angle(prec(p), p, succ(p) THEN radius(prec(p),
    p, succ(p)). This is also called the lexicographic maximum, and is the last
    entry of a list of (radius, angle) in lexicographical order.
    2a. If angle(prec(p), p, succ(p)) <= 90 degrees, then finish.
    2b. If not, remove p from set.
    """
    points = hull(points).tolist()
    if not_clockwise(points):
        points.reverse()
        if not_clockwise(points):
            raise Exception('Points are neither clockwise nor counterclockwise')
    POINTS = copy.deepcopy(points)
    removed = []
    i=0
    while True:
        angles = [_angle(_prec(p, points), p, _succ(p, points)) for p in points]
        circles = [_circle(_prec(p, points), p, _succ(p, points)) for p in points]
        radii = [c[0] for c in circles]
        lexord = np.lexsort((radii, angles)) #confusing as hell defaults...
        lexmax = lexord[-1]
        candidate = (_prec(points[lexmax], points),
                     points[lexmax],
                     _succ(points[lexmax], points))
        if angles[lexmax] <= PI/2.0:
            #print("Constrained by points: {}".format(candidate))
            return _circle(*candidate), points, removed, candidate
        else:
            try:
                removed.append((points.pop(lexmax), i))
            except IndexError:
                raise Exception("Construction of Minimum Bounding Circle failed!")
        i+=1

def _angle(p,q,r):
    """
    compute the positive angle formed by PQR
    """
    return np.abs(get_angle_between(Ray(q,p),Ray(q,r)))

def _prec(p,l):
    """
    retrieve the predecessor of p in list l
    """
    pos = l.index(p)
    if pos-1 < 0:
        return l[-1]
    else:
        return l[pos-1]

def _succ(p,l):
    """
    retrieve the successor of p in list l
    """
    pos = l.index(p)
    if pos+1 >= len(l):
        return l[0]
    else:
        return l[pos+1]

def _circle(p,q,r, dmetric=dist.euclidean):
    """
    Returns (radius, (center_x, center_y)) of the circumscribed circle by the
    triangle pqr.

    note, this does not assume that p!=q!=r
    """
    px,py = p
    qx,qy = q
    rx,ry = r
    if np.allclose(np.abs(_angle(p,q,r)), PI):
        radius = dmetric(p,r)/2.
        center_x = (px + rx)/2.
        center_y = (py + ry)/2.
    elif np.allclose(np.abs(_angle(p,q,r)), 0):
        radius = dmetric(p,q)/2.
        center_x = (px + qx)/2.
        center_y = (py + qy)/2.
    else:
        D = 2*(px*(qy - ry) + qx*(ry - py) + rx*(py - qy))
        center_x = ((px**2 + py**2)*(qy-ry) + (qx**2 + qy**2)*(ry-py)
                  + (rx**2 + ry**2)*(py-qy)) / float(D)
        center_y = ((px**2 + py**2)*(rx-qx) + (qx**2 + qy**2)*(px-rx)
                  + (rx**2 + ry**2)*(qx-px)) / float(D)
        radius = dmetric((center_x, center_y), p)
    return radius, (center_x, center_y)



__all__ = ['cached_property']
import functools


def cached_property(fun):
    """A memoize decorator for class properties.

    Adapted from: http://code.activestate.com/recipes/576563-cached-property/
    """
    @functools.wraps(fun)
    def get(self):
        try:
            return self._cache[fun]
        except AttributeError:
            self._cache = {}
        except KeyError:
            pass
        ret = self._cache[fun] = fun(self)
        return ret
    return property(get)



"""
Distance statistics for planar point patterns

"""
__author__ = "Serge Rey sjsrey@gmail.com"
__all__ = ['DStatistic', 'G', 'F', 'J', 'K', 'L', 'Envelopes', 'Genv', 'Fenv', 'Jenv', 'Kenv', 'Lenv']

from .process import PoissonPointProcess as csr
import numpy as np
from matplotlib import pyplot as plt


class DStatistic(object):
    """
    Abstract Base Class for distance statistics.

    Parameters
    ----------
    name       : string
                 Name of the function. ("G", "F", "J", "K" or "L")

    Attributes
    ----------
    d          : array
                 The distance domain sequence.

    """
    def __init__(self, name):
        self.name = name

    def plot(self, qq=False):
        """
        Plot the distance function

        Parameters
        ----------
        qq: Boolean
            If False the statistic is plotted against distance. If Frue, the
            quantile-quantile plot is generated, observed vs. CSR.
        """

        # assuming mpl
        x = self.d
        if qq:
            plt.plot(self.ev, self._stat)
            plt.plot(self.ev, self.ev)
        else:
            plt.plot(x, self._stat, label='{}'.format(self.name))
            plt.ylabel("{}(d)".format(self.name))
            plt.xlabel('d')
            plt.plot(x, self.ev, label='CSR')
            plt.title("{} distance function".format(self.name))


class G(DStatistic):
    """
    Estimates the nearest neighbor distance distribution function G for a
    point pattern.

    Parameters
    ----------
    pp         : :class:`.PointPattern`
                 Point Pattern instance.
    intervals  : int
                 The length of distance domain sequence.
    dmin       : float
                 The minimum of the distance domain.
    dmax       : float
                 The maximum of the distance domain.
    d          : sequence
                 The distance domain sequence.
                 If d is specified, intervals, dmin and dmax are ignored.

    Attributes
    ----------
    name       : string
                 Name of the function. ("G", "F", "J", "K" or "L")
    d          : array
                 The distance domain sequence.
    G          : array
                 The cumulative nearest neighbor distance distribution over d.

    Notes
    -----
    In the analysis of planar point processes, the estimate of :math:`G` is
    typically compared to the value expected from a completely spatial
    random (CSR) process given as:

    .. math::

            G(d) = 1 - e^{-\lambda \pi  d^2}

    where :math:`\lambda` is the intensity (points per unit area) of the point
    process and :math:`d` is distance.

    For a clustered pattern, the empirical function will be above the
    expectation, while for a uniform pattern the empirical function falls below
    the expectation.

    """

    def __init__(self, pp, intervals=10, dmin=0.0, dmax=None, d=None):
        res = _g(pp, intervals, dmin, dmax, d)
        self.d = res[:, 0]
        self.G = self._stat = res[:, 1]
        self.ev = 1 - np.exp(-pp.lambda_window * np.pi * self.d * self.d)
        self.pp = pp
        super(G, self).__init__(name="G")


class F(DStatistic):
    """
    Estimates the empty space distribution function for a point pattern: F(d).

    Parameters
    ----------
    pp         : :class:`.PointPattern`
                 Point Pattern instance.
    n          : int
                 Number of empty space points (random points).
    intervals  : int
                 The length of distance domain sequence.
    dmin       : float
                 The minimum of the distance domain.
    dmax       : float
                 The maximum of the distance domain.
    d          : sequence
                 The distance domain sequence.
                 If d is specified, intervals, dmin and dmax are ignored.

    Attributes
    ----------
    d          : array
                 The distance domain sequence.
    G          : array
                 The cumulative empty space nearest event distance distribution
                 over d.

    Notes
    -----
    In the analysis of planar point processes, the estimate of :math:`F` is
    typically compared to the value expected from a process that displays
    complete spatial randomness (CSR):

    .. math::

            F(d) = 1 - e^{-\lambda \pi  d^2}

    where :math:`\lambda` is the intensity (points per unit area) of the point
    process and :math:`d` is distance.

    The expectation is identical to the expectation for the :class:`G` function
    for a CSR process.  However, for a clustered pattern, the empirical G
    function will be below the expectation, while for a uniform pattern the
    empirical function falls above the expectation.

    """

    def __init__(self, pp, n=100, intervals=10, dmin=0.0, dmax=None, d=None):
        res = _f(pp, n, intervals, dmin, dmax, d)
        self.d = res[:, 0]
        self.F = self._stat = res[:, 1]
        self.ev = 1 - np.exp(-pp.lambda_window * np.pi * self.d * self.d)
        super(F, self).__init__(name="F")


class J(DStatistic):
    """
    Estimates the J function for a point pattern :cite:`VanLieshout1996`

    Parameters
    ----------
    pp         : :class:`.PointPattern`
                 Point Pattern instance.
    n          : int
                 Number of empty space points (random points).
    intervals  : int
                 The length of distance domain sequence.
    dmin       : float
                 The minimum of the distance domain.
    dmax       : float
                 The maximum of the distance domain.
    d          : sequence
                 The distance domain sequence.
                 If d is specified, intervals, dmin and dmax are ignored.

    Attributes
    ----------
    d          : array
                 The distance domain sequence.
    j          : array
                 F function over d.

    Notes
    -----

    The :math:`J` function is a ratio of the hazard functions defined for
    :math:`G` and :math:`F`:

    .. math::

            J(d) = \\frac{1-G(d) }{1-F(d)}

    where :math:`G(d)` is the nearest neighbor distance distribution function
    (see :class:`G`)
    and :math:`F(d)` is the empty space function (see :class:`F`).

    For a CSR process the J function equals 1. Empirical values larger than 1
    are indicative of uniformity, while values below 1 suggest clustering.


    """
    def __init__(self, pp, n=100, intervals=10, dmin=0.0, dmax=None, d=None):
        res = _j(pp, n, intervals, dmin, dmax, d)
        self.d = res[:, 0]
        self.j = self._stat = res[:, 1]
        self.ev = self.j / self.j
        super(J, self).__init__(name="J")


class K(DStatistic):
    """
    Estimates the  K function for a point pattern.

    Parameters
    ----------
    pp         : :class:`.PointPattern`
                 Point Pattern instance.
    intervals  : int
                 The length of distance domain sequence.
    dmin       : float
                 The minimum of the distance domain.
    dmax       : float
                 The maximum of the distance domain.
    d          : sequence
                 The distance domain sequence.
                 If d is specified, intervals, dmin and dmax are ignored.

    Attributes
    ----------
    d          : array
                 The distance domain sequence.
    j          : array
                 K function over d.

    """
    def __init__(self, pp, intervals=10, dmin=0.0, dmax=None, d=None):
        res = _k(pp, intervals, dmin, dmax, d)
        self.d = res[:, 0]
        self.k = self._stat = res[:, 1]
        self.ev = np.pi * self.d * self.d
        super(K, self).__init__(name="K")


class L(DStatistic):
    """
    Estimates the l function for a point pattern.

    Parameters
    ----------
    pp         : :class:`.PointPattern`
                 Point Pattern instance.
    intervals  : int
                 The length of distance domain sequence.
    dmin       : float
                 The minimum of the distance domain.
    dmax       : float
                 The maximum of the distance domain.
    d          : sequence
                 The distance domain sequence.
                 If d is specified, intervals, dmin and dmax are ignored.

    Attributes
    ----------
    d          : array
                 The distance domain sequence.
    l          : array
                 L function over d.
    """
    def __init__(self, pp, intervals=10, dmin=0.0, dmax=None, d=None):
        res = _l(pp, intervals, dmin, dmax, d)
        self.d = res[:, 0]
        self.l = self._stat = res[:, 1]
        super(L, self).__init__(name="L")

    def plot(self):
        # assuming mpl
        x = self.d
        plt.plot(x, self._stat, label='{}'.format(self.name))
        plt.ylabel("{}(d)".format(self.name))
        plt.xlabel('d')
        plt.title("{} distance function".format(self.name))


def _g(pp, intervals=10, dmin=0.0, dmax=None, d=None):
    """
    Estimate the nearest neighbor distances function G.

    Parameters
    ----------
    pp       : :class:`.PointPattern`
               Point Pattern instance.
    intevals : int
               Number of intervals to evaluate F over.
    dmin     : float
               Lower limit of distance range.
    dmax     : float
               Upper limit of distance range. If dmax is None, dmax will be set
               to maximum nearest neighor distance.
    d        : sequence
               The distance domain sequence. If d is specified, intervals, dmin
               and dmax are ignored.

    Returns
    -------
             : array
               A 2-dimensional numpy array of 2 columns. The first column is
               the distance domain sequence for the point pattern. The second
               column is the cumulative nearest neighbor distance distribution.

    Notes
    -----
    See :class:`G`.

    """
    if d is None:
        w = pp.max_nnd/intervals
        if dmax:
            w = dmax/intervals
        d = [w*i for i in range(intervals + 2)]
    cdf = [0] * len(d)
    for i, d_i in enumerate(d):
        smaller = [nndi for nndi in pp.nnd if nndi <= d_i]
        cdf[i] = len(smaller)*1./pp.n
    return np.vstack((d, cdf)).T


def _f(pp, n=100, intervals=10, dmin=0.0, dmax=None, d=None):
    """
    F empty space function.

    Parameters
    ----------
    pp       : :class:`.PointPattern`
               Point Pattern instance.
    n        : int
               Number of empty space points (random points).
    intevals : int
               Number of intervals to evaluate F over.
    dmin     : float
               Lower limit of distance range.
    dmax     : float
               Upper limit of distance range. If dmax is None, dmax will be set
               to maximum nearest neighor distance.
    d        : sequence
               The distance domain sequence. If d is specified, intervals, dmin
               and dmax are ignored.

    Returns
    -------
             : array
               A 2-dimensional numpy array of 2 columns. The first column is
               the distance domain sequence for the point pattern. The second
               column is corresponding F function.

    Notes
    -----
    See :class:`.F`

    """

    # get a csr pattern in window of pp
    c = csr(pp.window, n, 1, asPP=True).realizations[0]
    # for each point in csr pattern find the closest point in pp and the
    # associated distance
    nnids, nnds = pp.knn_other(c, k=1)

    if d is None:
        w = pp.max_nnd/intervals
        if dmax:
            w = dmax/intervals
        d = [w*i for i in range(intervals + 2)]
    cdf = [0] * len(d)

    for i, d_i in enumerate(d):
        smaller = [nndi for nndi in nnds if nndi <= d_i]
        cdf[i] = len(smaller)*1./n
    return np.vstack((d, cdf)).T


def _j(pp, n=100, intervals=10, dmin=0.0, dmax=None, d=None):
    """
    J function: Ratio of hazard functions for F and G.

    Parameters
    ----------
    pp       : :class:`.PointPattern`
               Point Pattern instance.
    n        : int
               Number of empty space points (random points).
    intevals : int
               Number of intervals to evaluate F over.
    dmin     : float
               Lower limit of distance range.
    dmax     : float
               Upper limit of distance range. If dmax is None, dmax will be set
               to maximum nearest neighor distance.
    d        : sequence
               The distance domain sequence. If d is specified, intervals, dmin
               and dmax are ignored.

    Returns
    -------
             : array
               A 2-dimensional numpy array of 2 columns. The first column is
               the distance domain sequence for the point pattern. The second
               column is corresponding J function.

    Notes
    -----
    See :class:`.J`

    """

    F = _f(pp, n, intervals=intervals, dmin=dmin, dmax=dmax, d=d)
    G = _g(pp, intervals=intervals, dmin=dmin, dmax=dmax, d=d)
    FC = 1 - F[:, 1]
    GC = 1 - G[:, 1]
    last_id = len(GC) + 1
    if np.any(FC == 0):
        last_id = np.where(FC == 0)[0][0]

    return np.vstack((F[:last_id, 0], GC[:last_id]/FC[:last_id])).T


def _k(pp, intervals=10, dmin=0.0, dmax=None, d=None):
    """
    Interevent K function.

    Parameters
    ----------
    pp       : :class:`.PointPattern`
               Point Pattern instance.
    n        : int
               Number of empty space points (random points).
    intevals : int
               Number of intervals to evaluate F over.
    dmin     : float
               Lower limit of distance range.
    dmax     : float
               Upper limit of distance range. If dmax is None, dmax will be set
               to length of bounding box diagonal.
    d        : sequence
               The distance domain sequence. If d is specified, intervals, dmin
               and dmax are ignored.

    Returns
    -------
    kcdf     : array
               A 2-dimensional numpy array of 2 columns. The first column is
               the distance domain sequence for the point pattern. The second
               column is corresponding K function.

    Notes
    -----
    See :class:`.K`

    """

    if d is None:
        # use length of bounding box diagonal as max distance
        bb = pp.mbb
        dbb = np.sqrt((bb[0]-bb[2])**2 + (bb[1]-bb[3])**2)
        w = dbb/intervals
        if dmax:
            w = dmax/intervals
        d = [w*i for i in range(intervals + 2)]
    den = pp.lambda_window * pp.n * 2.
    kcdf = np.asarray([(di, len(pp.tree.query_pairs(di))/den) for di in d])
    return kcdf


def _l(pp, intervals=10, dmin=0.0, dmax=None, d=None):
    """
    Interevent L function.

    Parameters
    ----------
    pp       : :class:`.PointPattern`
               Point Pattern instance.
    n        : int
               Number of empty space points (random points).
    intevals : int
               Number of intervals to evaluate F over.
    dmin     : float
               Lower limit of distance range.
    dmax     : float
               Upper limit of distance range. If dmax is None, dmax will be set
               to length of bounding box diagonal.
    d        : sequence
               The distance domain sequence. If d is specified, intervals, dmin
               and dmax are ignored.

    Returns
    -------
    kf       : array
               A 2-dimensional numpy array of 2 columns. The first column is
               the distance domain sequence for the point pattern. The second
               column is corresponding L function.

    Notes
    -----
    See :class:`.L`

    """

    kf = _k(pp, intervals, dmin, dmax, d)
    kf[:, 1] = np.sqrt(kf[:, 1] / np.pi) - kf[:, 0]
    return kf


class Envelopes(object):
    """
    Abstract base class for simulation envelopes.

    Parameters
    ----------
    pp          : :class:`.PointPattern`
                  Point Pattern instance.
    intervals   : int
                  The length of distance domain sequence. Default is 10.
    dmin        : float
                  The minimum of the distance domain.
    dmax        : float
                  The maximum of the distance domain.
    d           : sequence
                  The distance domain sequence.
                  If d is specified, intervals, dmin and dmax are ignored.
    pct         : float
                  1-alpha, alpha is the significance level. Default is 0.05,
                  1-alpha is the confidence level for the envelope.
    realizations: :class:`.PointProcess`
                  Point process instance with more than 1 realizations.

    Attributes
    ----------
    name        : string
                  Name of the function. ("G", "F", "J", "K" or "L")
    observed    : array
                  A 2-dimensional numpy array of 2 columns. The first column is
                  the distance domain sequence for the observed point pattern.
                  The second column is the specific function ("G", "F", "J",
                  "K" or "L") over the distance domain sequence for the
                  observed point pattern.
    low         : array
                  A 1-dimensional numpy array. Lower bound of the simulation
                  envelope.
    high        : array
                  A 1-dimensional numpy array. Higher bound of the simulation
                  envelope.
    mean        : array
                  A 1-dimensional numpy array. Mean values of the simulation
                  envelope.

    """
    def __init__(self, *args,  **kwargs):
        # setup arguments
        self.name = kwargs['name']

        # calculate observed function
        self.pp = args[0]
        self.observed = self.calc(*args, **kwargs)
        self.d = self.observed[:, 0]  # domain to be used in all realizations

        # do realizations
        self.mapper(kwargs['realizations'])

    def mapper(self, realizations):
        reals = realizations.realizations
        res = np.asarray([self.calc(reals[p]) for p in reals])

        # When calculating the J function for all the simulations, the length
        # of the returned interval domains might be different.

        if self.name == "J":
            res = []
            for p in reals:
                j = self.calc(reals[p])
                if j.shape[0] < self.d.shape[0]:
                    diff = self.d.shape[0]-j.shape[0]
                    for i in range(diff):
                        j = np.append(j, [[self.d[i+diff], np.inf]], axis=0)
                res.append(j)
            res = np.array(res)

        res = res[:, :, -1]
        res.sort(axis=0)
        nres = len(res)
        self.low = res[np.int(nres * self.pct/2.)]
        self.high = res[np.int(nres * (1-self.pct/2.))]
        self.mean = res.mean(axis=0)

    def calc(self, *args, **kwargs):
        print('implement in subclass')

    def plot(self):
        # assuming mpl
        x = self.d
        plt.plot(x, self.observed[:, 1], label='{}'.format(self.name))
        plt.plot(x, self.mean, 'g-.', label='CSR')
        plt.plot(x, self.low, 'r-.', label='LB')
        plt.plot(x, self.high, 'r-.', label="UB")
        plt.ylabel("{}(d)".format(self.name))
        plt.xlabel('d')
        plt.title("{} Simulation Envelopes".format(self.name))
        plt.legend(loc=0)


class Genv(Envelopes):
    """
    Simulation envelope for G function.

    Parameters
    ----------
    pp          : :class:`.PointPattern`
                  Point Pattern instance.
    intervals   : int
                  The length of distance domain sequence. Default is 10.
    dmin        : float
                  The minimum of the distance domain.
    dmax        : float
                  Upper limit of distance range. If dmax is None, dmax will be
                  set to maximum nearest neighbor distance.
    d           : sequence
                  The distance domain sequence.
                  If d is specified, intervals, dmin and dmax are ignored.
    pct         : float
                  1-alpha, alpha is the significance level. Default is 0.05,
                  which means 95% confidence level for the envelopes.
    realizations: :class:`.PointProcess`
                  Point process instance with more than 1 realizations.

    Attributes
    ----------
    name        : string
                  Name of the function. ("G", "F", "J", "K" or "L")
    observed    : array
                  A 2-dimensional numpy array of 2 columns. The first column is
                  the distance domain sequence for the observed point pattern.
                  The second column is cumulative nearest neighbor distance
                  distribution (G function) for the observed point pattern.
    low         : array
                  A 1-dimensional numpy array. Lower bound of the simulation
                  envelope.
    high        : array
                  A 1-dimensional numpy array. Higher bound of the simulation
                  envelope.
    mean        : array
                  A 1-dimensional numpy array. Mean values of the simulation
                  envelope.

    Examples
    --------
    .. plot::

       >>> import pysal.lib as ps
       >>> from pointpats import Genv, PoissonPointProcess, Window
       >>> from pysal.lib.cg import shapely_ext
       >>> va = ps.io.open(ps.examples.get_path("vautm17n.shp"))
       >>> polys = [shp for shp in va]
       >>> state = shapely_ext.cascaded_union(polys)
       >>> pp = PoissonPointProcess(Window(state.parts), 100, 1, asPP=True).realizations[0]
       >>> csrs = PoissonPointProcess(pp.window, 100, 100, asPP=True)
       >>> genv_bb = Genv(pp, realizations=csrs)
       >>> genv_bb.plot()

    """

    def __init__(self, pp, intervals=10, dmin=0.0, dmax=None, d=None, pct=0.05,
                 realizations=None):
        self.pp = pp
        self.intervals = intervals
        self.dmin = dmin
        self.dmax = dmax
        self.d = d
        self.pct = pct
        super(Genv, self).__init__(pp, realizations=realizations, name="G")

    def calc(self, *args, **kwargs):
        pp = args[0]
        return _g(pp, intervals=self.intervals, dmin=self.dmin, dmax=self.dmax,
                  d=self.d)


class Fenv(Envelopes):
    """
    Simulation envelope for F function.

    Parameters
    ----------
    pp          : :class:`.PointPattern`
                  Point Pattern instance.
    n           : int
                  Number of empty space points (random points).
    intervals   : int
                  The length of distance domain sequence. Default is 10.
    dmin        : float
                  The minimum of the distance domain.
    dmax        : float
                  Upper limit of distance range. If dmax is None, dmax will be
                  set to maximum nearest neighbor distance.
    d           : sequence
                  The distance domain sequence.
                  If d is specified, intervals, dmin and dmax are ignored.
    pct         : float
                  1-alpha, alpha is the significance level. Default is 0.05,
                  which means 95% confidence level for the envelopes.
    realizations: :class:`.PointProcess`
                  Point process instance with more than 1 realizations.

    Attributes
    ----------
    name        : string
                  Name of the function. ("G", "F", "J", "K" or "L")
    observed    : array
                  A 2-dimensional numpy array of 2 columns. The first column is
                  the distance domain sequence for the observed point pattern.
                  The second column is F function for the observed point
                  pattern.
    low         : array
                  A 1-dimensional numpy array. Lower bound of the simulation
                  envelope.
    high        : array
                  A 1-dimensional numpy array. Higher bound of the simulation
                  envelope.
    mean        : array
                  A 1-dimensional numpy array. Mean values of the simulation
                  envelope.

    Examples
    --------
    .. plot::

       >>> import pysal.lib as ps
       >>> from pysal.lib.cg import shapely_ext
       >>> from pointpats import PoissonPointProcess,Window,Fenv
       >>> va = ps.io.open(ps.examples.get_path("vautm17n.shp"))
       >>> polys = [shp for shp in va]
       >>> state = shapely_ext.cascaded_union(polys)
       >>> pp = PoissonPointProcess(Window(state.parts), 100, 1, asPP=True).realizations[0]
       >>> csrs = PoissonPointProcess(pp.window, 100, 100, asPP=True)
       >>> fenv = Fenv(pp, realizations=csrs)
       >>> fenv.plot()

    """
    def __init__(self, pp, n=100, intervals=10, dmin=0.0, dmax=None, d=None,
                 pct=0.05, realizations=None):
        self.pp = pp
        self.n = n
        self.intervals = intervals
        self.dmin = dmin
        self.dmax = dmax
        self.d = d
        self.pct = pct
        super(Fenv, self).__init__(pp, realizations=realizations, name="F")

    def calc(self, *args, **kwargs):
        pp = args[0]
        return _f(pp, self.n, intervals=self.intervals, dmin=self.dmin,
                  dmax=self.dmax, d=self.d)


class Jenv(Envelopes):
    """
    Simulation envelope for J function.

    Parameters
    ----------
    pp          : :class:`.PointPattern`
                  Point Pattern instance.
    n           : int
                  Number of empty space points (random points).
    intervals   : int
                  The length of distance domain sequence. Default is 10.
    dmin        : float
                  The minimum of the distance domain.
    dmax        : float
                  Upper limit of distance range. If dmax is None, dmax will be
                  set to maximum nearest neighbor distance.
    d           : sequence
                  The distance domain sequence.
                  If d is specified, intervals, dmin and dmax are ignored.
    pct         : float
                  1-alpha, alpha is the significance level. Default is 0.05,
                  which means 95% confidence level for the envelopes.
    realizations: :class:`.PointProcess`
                  Point process instance with more than 1 realizations.

    Attributes
    ----------
    name        : string
                  Name of the function. ("G", "F", "J", "K" or "L")
    observed    : array
                  A 2-dimensional numpy array of 2 columns. The first column is
                  the distance domain sequence for the observed point pattern.
                  The second column is J function for the observed point
                  pattern.
    low         : array
                  A 1-dimensional numpy array. Lower bound of the simulation
                  envelope.
    high        : array
                  A 1-dimensional numpy array. Higher bound of the simulation
                  envelope.
    mean        : array
                  A 1-dimensional numpy array. Mean values of the simulation
                  envelope.

    Examples
    --------
    .. plot::

       >>> import pysal.lib as ps
       >>> from pointpats import Jenv, PoissonPointProcess, Window
       >>> from pysal.lib.cg import shapely_ext
       >>> va = ps.io.open(ps.examples.get_path("vautm17n.shp"))
       >>> polys = [shp for shp in va]
       >>> state = shapely_ext.cascaded_union(polys)
       >>> pp = PoissonPointProcess(Window(state.parts), 100, 1, asPP=True).realizations[0]
       >>> csrs = PoissonPointProcess(pp.window, 100, 100, asPP=True)
       >>> jenv = Jenv(pp, realizations=csrs)
       >>> jenv.plot()

    """
    def __init__(self, pp, n=100, intervals=10, dmin=0.0, dmax=None, d=None,
                 pct=0.05, realizations=None):
        self.pp = pp
        self.n = n
        self.intervals = intervals
        self.dmin = dmin
        self.dmax = dmax
        self.d = d
        self.pct = pct
        super(Jenv, self).__init__(pp, realizations=realizations, name="J")

    def calc(self, *args, **kwargs):
        pp = args[0]
        return _j(pp, self.n, intervals=self.intervals, dmin=self.dmin,
                  dmax=self.dmax, d=self.d)


class Kenv(Envelopes):
    """
    Simulation envelope for K function.

    Parameters
    ----------
    pp          : :class:`.PointPattern`
                  Point Pattern instance.
    intervals   : int
                  The length of distance domain sequence. Default is 10.
    dmin        : float
                  The minimum of the distance domain.
    dmax        : float
                  Upper limit of distance range. If dmax is None, dmax will be
                  set to maximum nearest neighbor distance.
    d           : sequence
                  The distance domain sequence.
                  If d is specified, intervals, dmin and dmax are ignored.
    pct         : float
                  1-alpha, alpha is the significance level. Default is 0.05,
                  which means 95% confidence level for the envelope.
    realizations: :class:`.PointProcess`
                  Point process instance with more than 1 realizations.

    Attributes
    ----------
    name        : string
                  Name of the function. ("G", "F", "J", "K" or "L")
    observed    : array
                  A 2-dimensional numpy array of 2 columns. The first column is
                  the distance domain sequence for the observed point pattern.
                  The second column is K function for the observed point
                  pattern.
    low         : array
                  A 1-dimensional numpy array. Lower bound of the simulation
                  envelope.
    high        : array
                  A 1-dimensional numpy array. Higher bound of the simulation
                  envelope.
    mean        : array
                  A 1-dimensional numpy array. Mean values of the simulation
                  envelope.

    Examples
    --------
    .. plot::

       >>> import pysal.lib as ps
       >>> from pointpats import Kenv, PoissonPointProcess, Window
       >>> from pysal.lib.cg import shapely_ext
       >>> va = ps.io.open(ps.examples.get_path("vautm17n.shp"))
       >>> polys = [shp for shp in va]
       >>> state = shapely_ext.cascaded_union(polys)
       >>> pp = PoissonPointProcess(Window(state.parts), 100, 1, asPP=True).realizations[0]
       >>> csrs = PoissonPointProcess(pp.window, 100, 100, asPP=True)
       >>> kenv = Kenv(pp, realizations=csrs)
       >>> kenv.plot()

    """
    def __init__(self, pp, intervals=10, dmin=0.0, dmax=None, d=None,
                 pct=0.05, realizations=None):
        self.pp = pp
        self.intervals = intervals
        self.dmin = dmin
        self.dmax = dmax
        self.d = d
        self.pct = pct
        super(Kenv, self).__init__(pp, realizations=realizations, name="K")

    def calc(self, *args, **kwargs):
        pp = args[0]
        return _k(pp, intervals=self.intervals, dmin=self.dmin, dmax=self.dmax,
                  d=self.d)


class Lenv(Envelopes):
    """
    Simulation envelope for L function.

    Parameters
    ----------
    pp          : :class:`.PointPattern`
                  Point Pattern instance.
    intervals   : int
                  The length of distance domain sequence. Default is 10.
    dmin        : float
                  The minimum of the distance domain.
    dmax        : float
                  Upper limit of distance range. If dmax is None, dmax will be
                  set to maximum nearest neighbor distance.
    d           : sequence
                  The distance domain sequence.
                  If d is specified, intervals, dmin and dmax are ignored.
    pct         : float
                  1-alpha, alpha is the significance level. Default is 0.05,
                  which means 95% confidence level for the envelopes.
    realizations: :class:`.PointProcess`
                  Point process instance with more than 1 realizations.

    Attributes
    ----------
    name        : string
                  Name of the function. ("G", "F", "J", "K" or "L")
    observed    : array
                  A 2-dimensional numpy array of 2 columns. The first column is
                  the distance domain sequence for the observed point pattern.
                  The second column is L function for the observed point
                  pattern.
    low         : array
                  A 1-dimensional numpy array. Lower bound of the simulation
                  envelope.
    high        : array
                  A 1-dimensional numpy array. Higher bound of the simulation
                  envelope.
    mean        : array
                  A 1-dimensional numpy array. Mean values of the simulation
                  envelope.

    Examples
    --------
    .. plot::

       >>> import pysal.lib as ps
       >>> from pointpats import Lenv, PoissonPointProcess, Window
       >>> from pysal.lib.cg import shapely_ext
       >>> va = ps.io.open(ps.examples.get_path("vautm17n.shp"))
       >>> polys = [shp for shp in va]
       >>> state = shapely_ext.cascaded_union(polys)
       >>> pp = PoissonPointProcess(Window(state.parts), 100, 1, asPP=True).realizations[0]
       >>> csrs = PoissonPointProcess(pp.window, 100, 100, asPP=True)
       >>> lenv = Lenv(pp, realizations=csrs)
       >>> lenv.plot()

    """

    def __init__(self, pp, intervals=10, dmin=0.0, dmax=None, d=None,
                 pct=0.05, realizations=None):
        self.pp = pp
        self.intervals = intervals
        self.dmin = dmin
        self.dmax = dmax
        self.d = d
        self.pct = pct
        super(Lenv, self).__init__(pp, realizations=realizations, name="L")

    def calc(self, *args, **kwargs):
        pp = args[0]
        return _l(pp, intervals=self.intervals, dmin=self.dmin, dmax=self.dmax,
                  d=self.d)



import numpy as np

def explode(X):
    """
    Convert an (n,p) design matrix into an (n, n*p) local design matrix
    """
    n,p = X.shape
    Xs = X
    X = np.zeros((n, n*p))
    for i,row in enumerate(Xs):
        X[i, i*p:i*p+p] = row
    return X

def nexp(phi, pwds):
    """
    use the nexgative exponential distance weighting function on a matrix of
    pairwise distances to generate a spatial-dependence matrix for a svcp 
    """
    H = np.exp(- pwds / phi)
    return H 



import numpy as np


class NetworkBase(object):
    """Base object for performing network analysis on a spaghetti.Network
    object.
    
    Parameters
    ----------
    
    ntw : spaghetti.Network
        spaghetti Network object.
    
    pointpattern : spaghetti.network.PointPattern
        A spaghetti point pattern object.
    
    nsteps : int
        The number of steps at which the count of the nearest neighbors
        is computed.
    
    permutations : int
        The number of permutations to perform (default 99).
    
    threshold : float
        The level at which significance is computed.
        -- 0.5 would be 97.5% and 2.5%
    
    distribution : str
        The distribution from which random points are sampled
        -- uniform or poisson
    
    lowerbound : float
        The lower bound at which the function is computed. (Default 0).
    
    upperbound : float
        The upper bound at which the function is computed. Defaults to
        the maximum observed nearest neighbor distance.
    
    Attributes
    ----------
    
    sim : numpy.ndarray
        simulated distance matrix
    
    npts : int
        pointpattern.npoints
    
    xaxis : numpy.ndarray
        observed x-axis of values
    
    observed : numpy.ndarray
        observed y-axis of values
    
    """
    def __init__(self, ntw, pointpattern, nsteps=10, permutations=99,
                 threshold=0.5, distribution='poisson',
                 lowerbound=None, upperbound=None):
        self.ntw = ntw
        self.pointpattern = pointpattern
        self.nsteps = nsteps
        self.permutations = permutations
        self.threshold = threshold

        self.distribution = distribution
        self.validatedistribution()

        self.sim = np.empty((permutations, nsteps))
        self.npts = self.pointpattern.npoints

        self.lowerbound = lowerbound
        self.upperbound = upperbound

        # Compute Statistic.
        self.computeobserved()
        self.computepermutations()

        # Compute the envelope vectors.
        self.computeenvelope()


    def validatedistribution(self):
        """enusure statistical distribution is supported
        """
        valid_distributions = ['uniform', 'poisson']
        assert(self.distribution in valid_distributions),\
               "Distribution not in {}".format(valid_distributions)


    def computeenvelope(self):
        """compute upper and lower bounds of envelope
        """
        upper = 1.0 - self.threshold / 2.0
        lower = self.threshold / 2.0

        self.upperenvelope = np.nanmax(self.sim, axis=0) * upper
        self.lowerenvelope = np.nanmin(self.sim, axis=0) * lower


    def setbounds(self, nearest):
        """set upper and lower bounds
        """
        if self.lowerbound is None:
            self.lowerbound = 0
        if self.upperbound is None:
            self.upperbound = np.nanmax(nearest)


class NetworkG(NetworkBase):
    """Compute a network constrained G statistic. This requires the capability
    to compute a distance matrix between two point patterns. In this case one
    will be observed and one will be simulated.
    """
    def computeobserved(self):
        """compute the observed nearest
        """
        nearest = np.nanmin(self.ntw.allneighbordistances(self.pointpattern),
                            axis=1)
        self.setbounds(nearest)
        observedx, observedy = gfunction(nearest, self.lowerbound,
                                         self.upperbound, nsteps=self.nsteps)
        self.observed = observedy
        self.xaxis = observedx


    def computepermutations(self):
        """compute permutations of the nearest
        """
        for p in range(self.permutations):
            sim = self.ntw.simulate_observations(self.npts,
                                                distribution=self.distribution)
            nearest = np.nanmin(self.ntw.allneighbordistances(sim), axis=1)

            simx, simy = gfunction(nearest,
                                   self.lowerbound,
                                   self.upperbound,
                                   nsteps=self.nsteps)
            self.sim[p] = simy


class NetworkK(NetworkBase):
    """Compute a network constrained K statistic. This requires the capability
    to compute a distance matrix between two point patterns. In this case one
    will be observed and one will be simulated.
    
    Attributes
    ----------
    
    lam : float
        lambda value
    
    Notes
    -----
    
    Based on :cite:`Okabe2001`.
    
    """
    def computeobserved(self):
        """compute the observed nearest
        """
        nearest = self.ntw.allneighbordistances(self.pointpattern)
        self.setbounds(nearest)
        
        self.lam = self.npts / sum(self.ntw.edge_lengths.values())
        observedx, observedy = kfunction(nearest,
                                         self.upperbound,
                                         self.lam,
                                         nsteps=self.nsteps)
        self.observed = observedy
        self.xaxis = observedx


    def computepermutations(self):
        """compute permutations of the nearest
        """
        for p in range(self.permutations):
            sim = self.ntw.simulate_observations(self.npts,
                                                distribution=self.distribution)
            nearest = self.ntw.allneighbordistances(sim)
            
            simx, simy = kfunction(nearest,
                                   self.upperbound,
                                   self.lam,
                                   nsteps=self.nsteps)
            self.sim[p] = simy


class NetworkF(NetworkBase):
    """Compute a network constrained F statistic. This requires the capability
    to compute a distance matrix between two point patterns. In this case one
    will be observed and one will be simulated.
    
    Attributes
    ----------
    
    fsim : spaghetti.network.SimulatedPointPattern
        simulated point pattern of `self.npts` points
    
    """
    
    
    def computeobserved(self):
        """compute the observed nearest and simulated nearest
        """
        self.fsim = self.ntw.simulate_observations(self.npts)
        # Compute nearest neighbor distances from the simulated to the
        # observed.
        nearest = np.nanmin(self.ntw.allneighbordistances(self.fsim,
                                                          self.pointpattern),
                                                          axis=1)
        self.setbounds(nearest)
        # Generate a random distribution of points.
        observedx, observedy = ffunction(nearest, self.lowerbound,
                                         self.upperbound, nsteps=self.nsteps,
                                         npts=self.npts)
        self.observed = observedy
        self.xaxis = observedx


    def computepermutations(self):
        """compute permutations of the nearest
        """
        for p in range(self.permutations):
            sim = self.ntw.simulate_observations(self.npts,
                                                distribution=self.distribution)
            nearest = np.nanmin(self.ntw.allneighbordistances(sim, self.fsim),
                                axis=1)
            simx, simy = ffunction(nearest, self.lowerbound, self.upperbound,
                                   self.npts, nsteps=self.nsteps)
            self.sim[p] = simy


def gfunction(nearest, lowerbound, upperbound, nsteps=10):
    """Compute a G-Function

    Parameters
    ----------
    
    nearest : numpy.ndarray
        A vector of nearest neighbor distances.
    
    lowerbound : int or float
        The starting value of the sequence.
    
    upperbound : int or float
        The end value of the sequence.
    
    nsteps : int
        The number of distance bands. Default is 10. Must be non-negative.
    
    Returns
    -------
    
    x : numpy.ndarray
        x-axis of values
    
    y : numpy.ndarray
        y-axis of values
    
    """
    nobs = len(nearest)
    x = np.linspace(lowerbound, upperbound, nsteps)
    nearest = np.sort(nearest)
    
    y = np.empty(len(x))
    for i, r in enumerate(x):
        cnt = len(nearest[nearest <= r])
        if cnt > 0:
            g = cnt / float(nobs)
        else:
            g = 0
        y[i] = g
    return x, y


def kfunction(nearest, upperbound, intensity, nsteps=10):
    """Compute a K-Function

    Parameters
    ----------
    
    nearest : numpy.ndarray
        A vector of nearest neighbor distances.
    
    upperbound : int or float
        The end value of the sequence.
    
    intensity : float
        lambda value
    
    nsteps : int
        The number of distance bands. Default is 10. Must be non-negative.
    
    Returns
    -------
    
    x : numpy.ndarray
        x-axis of values
    
    y : numpy.ndarray
        y-axis of values
    
    """
    nobs = len(nearest)
    x = np.linspace(0, upperbound, nsteps)
    y = np.empty(len(x))
    
    for i, s in enumerate(x):
        y[i] = len(nearest[nearest <= s])
    y *= (intensity ** -1)
    return x, y


def ffunction(nearest, lowerbound, upperbound, npts, nsteps=10):
    """Compute an F-Function

    Parameters
    ----------
    
    nearest : numpy.ndarray
        A vector of nearest neighbor distances.
    
    lowerbound : int or float
        The starting value of the sequence.
    
    upperbound : int or float
        The end value of the sequence.
    
    npts : int
        pointpattern.npoints
    
    nsteps : int
        The number of distance bands. Default is 10. Must be non-negative.
    
    Returns
    -------
    
    x : numpy.ndarray
        x-axis of values
    
    y : numpy.ndarray
        y-axis of values
    
    """
    nobs = len(nearest)
    x = np.linspace(lowerbound, upperbound, nsteps)
    nearest = np.sort(nearest)
    y = np.empty(len(x))
    for i, r in enumerate(x):
        cnt = len(nearest[nearest <= r])
        if cnt > 0:
            g = cnt / float(npts)
        else:
            g = 0
        y[i] = g
    return x, y



__version__ = "2.0.1"
# __version__ has to be defined in the first line

from .classifiers import (Box_Plot, Equal_Interval, Fisher_Jenks,
                          Fisher_Jenks_Sampled, HeadTail_Breaks, Jenks_Caspall,
                          Jenks_Caspall_Forced, Jenks_Caspall_Sampled,
                          Max_P_Classifier, Maximum_Breaks, Natural_Breaks,
                          Quantiles, Percentiles, Std_Mean, User_Defined,
                          load_example, gadf, K_classifiers, CLASSIFIERS)




from __future__ import division
import scipy.stats as stats
import numpy as np
from warnings import warn
from ..steps import metropolis

def logp_phi(state, phi):
    """
    This is the log of the probability distribution in equation 7 of the
    appendix of the Wheeler and Calder (2010) paper on svcp
    """
    if phi < 0:
        return np.array([-np.inf])
    st = state
    # NOTE: I'm exploiting the following property for two square
    # matrices, H,T, of shapes n x n and p x p, respectively:
    # log(det(H kron T)) = log(det(H)^p * log(det(T)^n))
    # = log(det(H))*p + log(det(T))*n

    sgnH, logdetH = np.linalg.slogdet(st.H)
    sgnT, logdetT = np.linalg.slogdet(st.T)
    logdetH *= sgnH
    logdetT *= sgnT
    if any([x not in (-1,1) for x in [sgnH, sgnT]]):
        warn('Catastrophic loss of precision in np.linalg.slogdet of np.kron(st.H, st.T)')
    logdet = logdetH * st.p + logdetT * st.N
    Bmu = st.Betas - st.tiled_Mus
    kronHT_inv = st.kronHiTi #since inv of kronecker is kronecker of invs
    normal_kernel = np.dot(Bmu.T, np.dot(kronHT_inv, Bmu)) * -.5
    gamma_kernel = np.log(phi)*(st.Phi_shape0 - 1) + -1*st.Phi_rate0*phi
    return -.5*logdet + normal_kernel + gamma_kernel

def sample_phi(SVCP):
    """
    Sample phi, conditional on the state contained in the SVCP sampler

    Parameters
    ----------
    SVCP    :   sampler
                the execution context in which phi is to be sampled

    Returns
    --------
    None. works by sampling in place on SVCP. It updates:
    configs.phi.accepted OR configs.phi.rejected
    configs.phi.jump if tuning
    configs.phi.tuning if ending tuning
    state.Phi
    """
    state = SVCP.state
    cfg = SVCP.configs
    current = state.Phi
    try:
        #special proposals can be stored in configs
        proposal = cfg.Phi.proposal
    except KeyError:
        #if we don't have a proposal, take it to be a normal proposal
        proposal = stats.normal
        # and short circuit this assignment for later
        cfg.Phi.proposal = proposal
    new_val, accepted = metropolis(state, current, proposal,
                                   logp_phi,cfg.Phi.jump)

    if accepted:
        cfg.Phi.accepted += 1
    else:
        cfg.Phi.rejected += 1
    if cfg.tuning:
        acc = cfg.Phi.accepted
        rej = cfg.Phi.rejected
        ar = acc / (acc + rej)
        if ar < cfg.Phi.ar_low:
            cfg.Phi.jump *= cfg.Phi.adapt_step
        elif ar > cfg.Phi.ar_hi:
            cfg.Phi.jump /= cfg.Phi.adapt_step
        if SVCP.cycles >= cfg.Phi.max_tuning:
            cfg.tuning = False
    return new_val



"""
``splot.mapping``
=================

Provides Choropleth visualizations and mapping utilities.

Value-by-Alpha maps
-------------------

.. autosummary::
   :toctree: generated/

   value_by_alpha_cmap
   vba_choropleth
   vba_legend
   mapclassify_bin


Colormap utilities
------------------

.. autosummary::
   :toctree: generated/

   shift_colormap
   truncate_colormap

"""

from ._viz_value_by_alpha_mpl import (value_by_alpha_cmap,
                                      vba_choropleth,
                                      vba_legend,
                                      mapclassify_bin)

from ._viz_utils import (shift_colormap,
                         truncate_colormap)



"""
``splot.giddy``
===============

Provides visualisations for the Geospatial Distribution Dynamics - `giddy` module.
`giddy` provides a tool for space–time analytics that consider the role of space in the evolution of distributions over time.

Directional LISA analytics
--------------------------

.. autosummary::
   :toctree: generated/

   dynamic_lisa_heatmap
   dynamic_lisa_rose
   dynamic_lisa_vectors
   dynamic_lisa_composite
   dynamic_lisa_composite_explore

"""

from ._viz_giddy_mpl import (dynamic_lisa_heatmap,
                             dynamic_lisa_rose,
                             dynamic_lisa_vectors,
                             dynamic_lisa_composite,
                             dynamic_lisa_composite_explore)



from __future__ import division
import numpy as _np
import pandas as _pd
from .utils import thru_op as _thru_op
from .abstracts import Trace as _Trace, Hashmap as _Hashmap, _maybe_hashmap, _copy_hashmaps
from warnings import warn as _Warn
import copy as _copy

__all__ = ['summarize', 'mcse', 'psrf', 'geweke', 'hpd_interval', 'effective_size']

try:
    from rpy2.rinterface import RRuntimeError
    from rpy2.robjects.packages import importr
    from rpy2.robjects.numpy2ri import numpy2ri
    import rpy2.robjects as ro
    ro.conversion.py2ri = numpy2ri
    _coda = importr('coda')
    _HAS_CODA = True
    _HAS_RPY2 = True
except (ImportError, LookupError):
    _HAS_CODA = False
    _HAS_RPY2 = False
except RRuntimeError:
    _HAS_CODA = False
    _HAS_RPY2 = True

#############
# Summarize #
#############

def summarize(model = None, trace = None, chain=None, varnames=None,
              level=0):
    """
    Summarize a trace object, providing its mean, median, HPD,
    standard deviation, and effective size.

    Arguments
    ---------
    model   :   Any model object.
                must have an attached `trace` attribute. Takes precedence over
                `trace` and `chain` arguments.
    trace   :   abstracts.Trace
                a trace object containing data to compute the diagnostic
    chain   :   np.ndarray
                an array indexed by (m,n[,p]) containing m parallel runs of n samples
                of p covariates.
    varnames:   str or list of str
                set of variates to extract from the model or trace to to compute the
                statistic.
    level   :   int
                ordered in terms of how much information reduction occurs. a level 0 summary
                provides the output for each chain. A level 1 summary provides output
                grouped over all chains.
    """
    trace = _resolve_to_trace(model, trace, chain, varnames)
    dfs = trace.to_df()
    if isinstance(dfs, list):
        multi_index = ['Chain_{}'.format(i) for i in range(len(dfs))]
        df = _pd.concat(dfs, axis=1, keys=multi_index)
    else:
        df = _pd.concat((dfs,), axis=1, keys=['Chain_0'])
    df = df.describe().T[['count', 'mean', '50%', 'std']]
    HPDs = hpd_interval(trace=trace, p=.95)
    if _HAS_CODA:
        ESS = effective_size(trace=trace, use_R=True)
    else:
        _Warn('Computing effective sample size may take a while due to statsmodels.tsa.AR.',
             stacklevel=2)
        ESS = effective_size(trace=trace, use_R=False)
    flattened_HPDs = []
    flattened_ESSs = []
    if isinstance(HPDs, dict):
        HPDs = [HPDs]
    if isinstance(ESS, dict):
        ESS = [ESS]
    for i_chain, chain in enumerate(HPDs):
        this_HPD = dict()
        this_ESS = dict()
        for key, val in chain.items():
            if isinstance(val, list):
                for i, hpd_tuple in enumerate(val):
                    name = '{}_{}'.format(key, i)
                    this_HPD.update({name: hpd_tuple})
                    this_ESS.update({name: ESS[i_chain][key][i]})
            else:
                this_HPD.update({key: val})
                this_ESS.update({key: ESS[i_chain][key]})
        flattened_HPDs.append(this_HPD)
        flattened_ESSs.append(this_ESS)
    df['HPD_low'] = None
    df['HPD_high'] = None
    df['N_effective'] = None
    for i, this_chain_HPD in enumerate(flattened_HPDs):
        this_chain_ESS = flattened_ESSs[i]
        outer_key = 'Chain_{}'.format(i)
        keys = [(outer_key, inner_key) for inner_key in this_chain_HPD.keys()]
        lows, highs = zip(*[this_chain_HPD[key[-1]] for key in keys])
        n_eff = [this_chain_ESS[key[-1]] for key in keys]
        df.ix[keys, 'HPD_low'] = lows
        df.ix[keys, 'HPD_high'] = highs
        df.ix[keys, 'N_effective'] = n_eff
    df['median'] = df['50%']
    df['N_iters'] = df['count'].apply(int)
    df['N_effective'] = df['N_effective'].apply(round)
    df.drop('count', axis=1, inplace=True)
    df['AR_loss'] = (df['N_iters'] - df['N_effective'])/df['N_iters']
    df = df[['mean', 'HPD_low', 'median', 'HPD_high', 'std',
             'N_iters', 'N_effective', 'AR_loss']]
    if level > 0:
        df = df.unstack()
        grand_mean = df['mean'].mean(axis=0)
        lowest_HPD = df['HPD_low'].min(axis=0)
        grand_median = df['median'].median(axis=0)
        highest_HPD = df['HPD_high'].max(axis=0)
        std = df['std'].mean(axis=0)
        neff = df['N_effective'].sum(axis=0)
        N = df['N_iters'].sum(axis=0)
        df = _pd.concat([grand_mean, lowest_HPD, grand_median,
                        highest_HPD, std, N, neff], axis=1)
        df.columns = ['grand_mean', 'min_HPD', 'grand_median', 'max_HPD', 'std',
                      'sum(N_iters)', 'sum(N_effective)']
    return df

#####################################
# Potential Scale Reduction Factors #
#####################################


def _gelman_rubin(chain):
    """
    Computes the original potential scale reduction factor from the
     1992 paper by Gelman and Rubin:

    \sqrt{\frac{\hat{V}}{W} * \frac{dof}{dof-2}}

    where \hat{V} is a corrected estimate of the chain variance, composed of within- and between-chain variance components, W,B:

    V_hat = W*(1-1/n) + B/n + B/(n*m)

    and the degrees of freedom terms are:
    dof = 2 * V_hat**2 / (Var(W) + Var(B) + 2Cov(W,B)).

    The equations of the variance and covariance are drawn directly from the original paper. This implementation should come close to the implementation in the R CODA package, which computes the same normalization factor.

    If the chain is multivariate, it computes the statistic for each element of the multivariate chain.
    """
    m,n = chain.shape[0:2]
    rest = chain.shape[2:]
    if len(rest) == 0:
        chain_vars = chain.var(axis=1, ddof=1)
        chain_means = chain.mean(axis=1)
        grand_mean = chain.mean()

        W = _np.mean(chain_vars)
        B = _np.var(chain_means, ddof=1)*n
        sigma2_hat = W*(1-(1/n)) + B/n
        V_hat = sigma2_hat + B/(n*m)
        t_scale = _np.sqrt(V_hat)

        #  not sure if the chain.var(axis=1, ddof=1).var() is right.
        var_W = (1-(1/n))**2 * chain_vars.var(ddof=1) / m
        var_B = ((m+1)/(m*n))**2 * (2*B**2) / (m-1)
        cov_s2xbar2 = _np.cov(chain_vars, chain_means**2, ddof=1)[0, 1]
        cov_s2xbarmu = 2 * grand_mean * _np.cov(chain_vars, chain_means, ddof=1)[0, 1]
        cov_WB = (m+1)*(n-1)/(m*n**2)*(m/n)*(cov_s2xbar2 - cov_s2xbarmu)

        t_dof = 2 * V_hat**2 / (var_W + var_B + 2 * cov_WB)

        psrf = _np.sqrt((V_hat / W) * t_dof/(t_dof - 2))
        return psrf
    else:
        return [_gelman_rubin(ch.T) for ch in chain.T]


def _brooks_gelman_rubin(chain):
    """
    Computes the Brooks and Gelman psrf in equation 1.1, pg. 437 of the Brooks and Gelman article.

    This form is:
    (n_chains + 1) / n_chains * (sigma2_hat / W) - (n-1)/(nm)
    where Sigma2_hat is the unbiased estimator for aggregate variance:
    (n-1)/n * W + B/n

    If the chain is multivariate, this computes the univariate version over all elements of the chain.
    """
    m,n = chain.shape[0:2]
    rest = chain.shape[2:]
    if len(rest) == 0:
        chain_vars = chain.var(axis=1, ddof=1)
        chain_means = chain.mean(axis=1)
        grand_mean = chain.mean()

        W = _np.mean(chain_vars)
        B = _np.var(chain_means, ddof=1)*n

        sigma2_hat = ((n-1)/n) * W + B/n
        Rhat = (m+1)/m * (sigma2_hat / W) - ((n-1)/(m*n))
        return _np.sqrt(Rhat)
    else:
        return [_brooks_gelman_rubin(ch.T) for ch in chain.T]

_psrf = dict([('brooks', _brooks_gelman_rubin), ('original',_gelman_rubin)])

def psrf(model = None, trace=None, chain=None, autoburnin=True,
         varnames=None, method='brooks'):
    """
    Wrapper to compute the potential scale reduction factor for
    a trace object or an arbitrary chain from a MCMC sample.

    Arguments
    ----------
    trace       :  Trace
                   A trace object that contains more-than-one chain.
    chain       :  np.ndarray
                   An array with at least two dimensions where the indices are:
                   (m,n[, k]), where m is the number of traces and n is the number of iterations. If the parameter is k-variate, then the trailing dimension must be k.
    autoburnin  :  boolean
                   a flat denoting whether to automatically slice the chain at its midpoint, computing the psrf for only the second half.
    varnames    :  string or list of strings
                   collection of the names of variables to compute the psrf.
    method      :  string
                   the psrf statistic to be used. Recognized options:
                   - 'brooks' (default): the 1998 Brooks-Gelman-Rubin psrf
                   - 'original': the 1992 Gelman-Rubin psrf
    """
    if model is not None:
        trace = model.trace
    if trace is not None and varnames is None:
        varnames = trace.varnames
    elif chain is not None and varnames is None:
        varnames = ['parameter']
    elif chain is not None and varnames is not None:
        try:
            assert len(varnames) == 1
        except AssertionError:
            raise UserWarning('Multiple chains outside of a trace '
                              'are not currently supported')
    out = dict()
    for param in varnames:
        if chain is not None:
            this_chain = chain
            m,n = chain.shape[0:2]
            rest = chain.shape[2:]
        else:
            this_chain = trace[param]
            m,n = this_chain.shape[0:2]
            rest = this_chain.shape[2:]
        this_chain = this_chain[:,-n//2:,]
        out.update({param:_psrf[method](this_chain)})
    return out

######################
# Geweke Diagnostics #
######################

def geweke(model = None, trace=None, chain=None,
           drop_frac=.1, hold_frac=.5, n_bins=50,
           varnames=None, variance_method='ar', **ar_kw):
    """
    This computes the plotting version of Geweke's diagnostic for a given trace. The iterative version is due to Brooks. This implementation mirrors that in the `R` `coda` package.

    Arguments
    ----------
    model   :   Any model object.
                must have an attached `trace` attribute. Takes precedence over
                `trace` and `chain` arguments.
    trace   :   abstracts.Trace
                a trace object containing data to compute the diagnostic
    chain   :   np.ndarray
                an array indexed by (m,n[,p]) containing m parallel runs of n samples of p covariates.
    drop_frac:  float
                the number of observations to drop each step from the first (1-`keep_frac`)% of the chain
    keep_frac:  float
                the comparison group of observations used to compare the to
                the bins over the (1 - keep_frac)% of the chain
    n_bins  :   int
                number of bins to divide the first (1 - keep_frac)% of the chain
                into.
    varnames:   string or list of strings
                name or list of names of parameters to which the diagnostic should be applied.
    variance_method: str
                name of the variance method to be used. The default, `ar0`, uses the spectral density at lag 0, which is also used in CODA. This corrects for serial correlation in the variance estimate. The alternative, `naive`, is simply the sample variance.
    ar_kw   :   dict/keyword arguments
                If provided, must contain `spec_kw` and `fit_kw`. `spec_kw` is a dictionary of keyword arguments passed to the statsmodels AR class, and `fit_kw` is a dictionary of arguments passed to the subsequent AR.fit() call.
    """
    trace = _resolve_to_trace(model, trace, chain, varnames)
    varnames = trace.varnames
    variance_function = _geweke_variance[variance_method]
    all_stats = []
    for i, chain in enumerate(trace.chains):
        all_stats.append(dict())
        for var in varnames:
            data = _np.squeeze(trace[i,var])
            if data.ndim > 1:
                n,p = data.shape[0:2]
                rest = data.shape[2:0]
                if len(rest) == 0:
                    data = data.T
                elif len(rest) == 1:
                    data = data.reshape(n,p*rest[0]).T
                else:
                    raise Exception('Parameter "{}" shape not understood.'                  ' Please extract, shape it, and pass '
                                    ' as its own chain. '.format(var))
            else:
                data = data.reshape(1,-1)
            stats = [_geweke_vector(datum, drop_frac, hold_frac, n_bins=n_bins,
                                    varfunc=variance_function)
                    for datum in data]
            if len(stats) > 1:
                results = {"{}_{}".format(var, i):stat for i,stat in
                            enumerate(stats)}
            else:
                results = {var:stats[0]}
            all_stats[i].update(results)
    return all_stats

def _geweke_map(model = None, trace=None, chain=None,
           drop_frac=.1, hold_frac=.5, n_bins=50,
           varnames=None, variance_method='ar', **ar_kw):
    trace = _resolve_to_trace(model, trace, chain, varnames)
    stats = trace.map(_geweke_vector, drop=drop_frac, hold=hold_frac, n_bins=n_bins,
                      varfunc=_geweke_variance[variance_method])
    return stats

def _geweke_vector(data, drop, hold, n_bins, **kw):
    """
    Compute a vector of geweke statistics over the `data` vector. This proceeds like the `R` `CODA` package's geweke statistic. The first half of the data vector is split into `n_bins` segments. Then, the Geweke statistic is repeatedly computed over subsets of the data where a bin is dropped each step. This results in `n_bins` statistics.
    """
    in_play = (len(data)-1)//2
    to_drop = _np.linspace(0, in_play, num=n_bins).astype(int)
    return _np.squeeze([_geweke_statistic(data[drop_idx:], drop, hold, **kw)
                       for drop_idx in to_drop])

def _geweke_statistic(data, drop, hold, varfunc=None):
    """
    Compute a single geweke statistic, defining sets A, B:

    mean_A - mean_B / (var(A) + var(B))**.5

    where A is the first `drop` % of the `data` vector, B is the last `hold` % of the data vector.

    the variance function, `varfunc`, is the spectral density estimate of the variance.
    """
    if varfunc is None:
        varfunc = _spectrum0_ar
    hold_start = _np.floor((len(data)-1) * hold).astype(int)
    bin_width = _np.ceil((len(data)-1)*drop).astype(int)

    drop_data = data[:bin_width]
    hold_data = data[hold_start:]

    drop_mean = drop_data.mean()
    drop_var = varfunc(drop_data)
    n_drop = len(drop_data)

    hold_mean = hold_data.mean()
    hold_var = varfunc(hold_data)
    n_hold = len(hold_data)

    return ((drop_mean - hold_mean) / _np.sqrt((drop_var / n_drop)
                                            +(hold_var / n_hold)))

def _naive_var(data, *_, **__):
    """
    Naive variance computation of a time `x`, ignoring dependence between the
    variance within different windows
    """
    return _np.var(data, ddof=1)

def _spectrum0_ar(data, spec_kw=dict(), fit_kw=dict()):
    """
    The corrected spectral density estimate of time series variance,
    as applied in CODA. Written to replicate R, so defaults change.
    Note: this is very slow when there is a lot of data.
    """
    try:
        from statsmodels.api import tsa
    except ImportError:
        raise ImportError('Statsmodels is required to use the AR(0) '
                           ' spectral density estimate of the variance.')
    if fit_kw == dict():
        fit_kw['ic']='aic'
        N = len(data)
        # R uses the smaller of N-1 and 10*log10(N). We should replicate that.
        maxlag = N-1 if N-1 <= 10*_np.log10(N) else 10*_np.log(N)
        fit_kw['maxlag'] = int(_np.ceil(maxlag))
    ARM = tsa.AR(data, **spec_kw).fit(**fit_kw)
    alphas = ARM.params[1:]
    return ARM.sigma2 / (1 - alphas.sum())**2

_geweke_variance = dict()
_geweke_variance['ar'] = _spectrum0_ar
_geweke_variance['naive'] = _naive_var

##################
# Effective Size #
##################

def effective_size(model=None, trace=None, chain=None, varnames=None,
                    use_R=False):
    """
    Compute the effective size of a trace, accounting for serial autocorrelation.
    This statistic is:

    N * var(x)/spectral0(x)

    where spectral0(x) is the spectral density of x at lag 0, an
    autocorrelation-adjusted estimate of the sequence variance.

    NOTE: the backend argument defaults to estimating the effective_size in
    python. But, the statsmodels.tsa.AR required for the spectral density
    estimate is *slow* for large chains. If you have a properly configured R
    installation with the python package `rpy2` and the R package `coda` installed,
    you can opt to pass through to CODA by passing `use_R=True`.

    Arguments
    ----------
    model   :   Any model object.
                must have an attached `trace` attribute. Takes precedence over
                `trace` and `chain` arguments.
    trace   :   abstracts.Trace
                a trace object containing data to compute the diagnostic
    chain   :   np.ndarray
                an array indexed by (m,n[,p]) containing m parallel runs of n samples
                of p covariates.
    varnames:   str or list of str
                set of variates to extract from the model or trace to to compute the
                statistic.
    use_R   :   bool (default: False)
                option to drop the computation of the effective size down to R's CODA
                package. Requires: rpy2, working R installation, CODA R package
    """
    trace = _resolve_to_trace(model, trace, chain, varnames)
    stats = trace.map(_effective_size, use_R=use_R)
    return stats if len(stats) > 1 else stats[0]

def _effective_size(x, use_R=False):
    """
    Compute the effective size from a given flat array

    Arguments
    -----------
    x       :   np.ndarray
                flat vector of values to compute the effective sample size

    use_R   :   bool
                option to use rpy2+CODA or pure python implementation. If False,
                the effective size computation may be unbearably slow on large data,
                due to slow AR fitting in statsmodels.
    """
    if use_R:
        if _HAS_RPY2 and _HAS_CODA:
            return _coda.effectiveSize(x)[0]
        elif _HAS_RPY2 and not _HAS_CODA:
            raise ImportError("No module named 'coda' in R.")
        else:
            raise ImportError("No module named 'rpy2'")
    else:
        spec = _spectrum0_ar(x)
        if spec == 0:
            return 0
        loss_factor = _np.var(x, ddof=1)/spec
        return len(x) * loss_factor

#############################
# Highest Posterior Density #
#############################

def hpd_interval(model = None,  trace = None,  chain = None,  varnames = None,  p=.95):
    """

    Parameters
    ----------
    model   :   Any model object.
                must have an attached `trace` attribute. Takes precedence over
                `trace` and `chain` arguments.
    trace   :   abstracts.Trace
                a trace object containing data to compute the diagnostic
    chain   :   np.ndarray
                an array indexed by (m,n[,p]) containing m parallel runs of n samples
                of p covariates.
    varnames:   str or list of str
                set of variates to extract from the model or trace to to compute the
                statistic.
    p       :   float
                percent of highest density to extract

    Returns
    -------
    hashmap of results, where each result is {'varname':(low, hi)}

    """
    trace = _resolve_to_trace(model, trace, chain, varnames)
    stats = trace.map(_hpd_interval, p=p)
    return stats if len(stats) > 1 else stats[0]

def _hpd_interval(data, p=.95):
    """

    Parameters
    ----------
    data    :   numpy.ndarray
                data to compute the hpd
    p       :   float
                percent of highest density to extract

    Returns
    -------
    tuple of (low,hi) boundaries of the highest posterior fraction.

    """
    data = _np.sort(data)
    N = len(data)
    N_in = int(_np.ceil(N*p))
    head = _np.arange(0,N-N_in)
    tail = head+N_in
    pivot = _np.argmin(data[tail] - data[head])
    return data[pivot], data[pivot+N_in]

def point_estimates(model=None, trace=None, chain=None, 
                    burnin=0,thin=1,
                    varnames=None, statistic=_np.median):
    """
    Get a point estimate for the posterior
    """
    trace = _resolve_to_trace(model, trace, chain, varnames)
    trace = _resolve_to_trace(None, trace[burnin::thin], None, varnames)
    return trace.map(statistic)


############################################
# Markov Chain Monte Carlo Standard Errors #
############################################

def mcse(model = None, trace=None, chain = None, varnames = None,
           rescale=2, method='bartlett', N_chunks=None, transform=_thru_op):
    """

    Parameters
    ----------
    model   :   Any model object.
                must have an attached `trace` attribute. Takes precedence over
                `trace` and `chain` arguments.
    trace   :   abstracts.Trace
                a trace object containing data to compute the diagnostic
    chain   :   np.ndarray
                an array indexed by (m,n[,p]) containing m parallel runs of n samples
                of p covariates.
    varnames:   str or list of str
                set of variates to extract from the model or trace to to compute the
                statistic.
    rescale :   real positive number
                governs the shrinkage/reduction of the relevant data
    method  :   str (default: 'bartlett')
                string describing method used to compute the standard errors.
                Supported options:
                    - 'bm' : batch means, reduces chain by the mean of each `N_chunks` block
                    - 'obm': overlapping batch means, reduces chain by the mean of `N_chunks` rolling blocks
                    - 'tukey': weighted reduction using a Tukey window (see np.tukey)
                    - 'hanning': same as 'tukey'
                    - 'bartlett': weighted reduction using a Bartlett window (see np.bartlett)
    transform:  callable
                function or callable class that consumes data and returns it in an identical shape. Used
                if the values of interest are a transformation of the given parameter.
    Returns
    -------
    hashmap or list of hashmaps that contain the standard errors of the given chain.
    """
    trace = _resolve_to_trace(model, trace, chain, varnames)
    out = trace.map(_mcse, rescale=rescale, method=method, N_chunks=N_chunks, transform=transform)
    return out if len(out) > 1 else out[0]

def _mcse(x, rescale=2, N_chunks=None, method='bm', transform=_thru_op):
    if rescale is None and N_chunks is None:
        raise ValueError("Either 'rescale' or 'N_chunks' must be supplied.")
    elif rescale is not None and N_chunks is None:
        size = len(x)
        if 0 < rescale < 1:
            rescale = 1/rescale
        chunk_size = _np.floor(size**(1.0/rescale))
        N_chunks = _np.floor(size / chunk_size)
    elif N_chunks is not None:
        pass
    else:
        raise Exception("Options 'rescale' and 'N_chunks' were not resolved successfully!")

    try:
        method = _mcse_dispatch[method]
    except KeyError:
        raise KeyError("Supported methods are: 'bm', 'obm', 'bartlett', 'tukey'")
    return method(x, int(N_chunks), transform = transform)

def _mcse_bm(x, N_chunks, transform = _thru_op):
    """
    Compute a Markov Chain Monte Carlo Standard Error
     using raw batch means

    Parameters
    ----------
    x           :   numpy.ndarray
    N_chunks    :   int
    transform   :   callable

    Returns
    -------
    float containing the standard error of x
    """
    N = len(x)
    chunk_size = _np.floor(N / N_chunks).astype(int)
    y = _np.asarray([transform(split).mean() for split in _np.array_split(x, N_chunks)])
    mean = transform(x).mean()
    variance = chunk_size * ((y - mean)**2).sum() / (N_chunks - 1) #isn' this chunk_size * (x.var(ddof=1)?)
    return _np.sqrt(variance / N)

def _mcse_obm(x, N_chunks, transform = _thru_op):
    """
    Compute a Markov Chain Monte Carlo Standard Error
     using overlapping batch means

    Parameters
    ----------
    x           :   numpy.ndarray
    N_chunks    :   int
    transform   :   callable

    Returns
    -------
    float containing the standard error of x
    """
    N = len(x)
    a = N - N_chunks + 1
    chunk_size = _np.floor(N / N_chunks).astype(int)
    y = _pd.Series(x).rolling(chunk_size).apply(lambda vec: _np.mean(transform(vec)))
    y = y[~_np.isnan(y)]
    mean = transform(x).mean()
    variance = N * chunk_size * ((y - mean)**2).sum() / (a -1) / a
    return _np.sqrt(variance / N)

def _mcse_bartlett(x, N_chunks, transform = _thru_op):
    """
    Compute a Markov Chain Monte Carlo Standard Error
     using a Bartlett window

    Parameters
    ----------
    x           :   numpy.ndarray
    N_chunks    :   int
    transform   :   callable

    Returns
    -------
    float containing the standard error of x
    """
    N = len(x)
    chunk_size = _np.floor(N / N_chunks).astype(int)
    brange = _np.arange(1, chunk_size+1)
    alpha = (1 - brange / chunk_size) * (1 - brange / N)
    mean = transform(x).mean()
    diffs = ((x[0:(N-i)] - mean) * (x[i:N] - mean) for i in range(chunk_size+1))
    R = _np.asarray([diff.mean() for diff in diffs])
    variance = R[0] + 2 * (alpha * R[1:]).sum()
    return _np.sqrt(variance / N)

def _mcse_hanning(x, N_chunks, transform = _thru_op):
    """
    Compute a Markov Chain Monte Carlo Standard Error
     using a Tukey window

    Parameters
    ----------
    x           :   numpy.ndarray
    N_chunks    :   int
    transform   :   callable

    Returns
    -------
    float containing the standard error of x
    """
    N = len(x)
    chunk_size = _np.floor(N / N_chunks).astype(int)
    brange = _np.arange(1, chunk_size+1)
    alpha = (1 + _np.cos(_np.pi * brange / chunk_size)) / 2 * (1 - brange / N)
    mean = transform(x).mean()
    diffs = ((x[0:(N-i)] - mean) * (x[i:N] - mean) for i in range(chunk_size+1))
    R = _np.asarray([diff.mean() for diff in diffs])
    variance = R[0] + 2 * (alpha * R[1:]).sum()
    return _np.sqrt(variance / N)

_mcse_dispatch = dict(
    tukey = _mcse_hanning,
    hanning = _mcse_hanning,
    bartlett = _mcse_bartlett,
    obm = _mcse_obm,
    bm = _mcse_bm
)

#############
# Utilities #
#############

def _resolve_to_trace(model, trace, chain, varnames):
    """
    Resolve a collection of information down to a trace. This reduces the
    passed arguments to a trace that can be used for analysis based on names
    in varnames.

    If `trace` is passed, it is subset according to `varnames`, and a copy returned.
    It takes precedence.
    Otherwise, if `model` is passed, its traces are taken.
    Finally, if `chain` is passed, a trace is constructed to structure the chain.
    In all cases, if `varnames` is passed, it is used to name or subset the given data.

    """
    n_passed = sum([model is not None, trace is not None, chain is not None])
    if n_passed > 1:
        raise Exception('Only one of `model`, `trace`, or `chain` '
                        ' may be passed.')
    if isinstance(varnames, str):
        varnames = [varnames]
    if trace is not None:
        if isinstance(trace, _Trace):
            if varnames is not None:
                return trace.drop([var for var in trace.varnames
                                    if var not in varnames], inplace=False)
            else:
                return _copy.deepcopy(trace)
        else:
            return _Trace(*_copy_hashmaps(_maybe_hashmap(trace)[0]))

    if model is not None:
        return _resolve_to_trace(model=None, trace=model.trace,
                                 chain=None, varnames=varnames)
    if chain is not None:
        if chain.ndim > 2:
            m, n = chain.shape[0:2]
            rest = chain.shape[2:]
            new_p = _np.multiply(*rest)
            chain = chain.reshape(m, n, new_p)
        elif chain.ndim == 1:
            if varnames is None:
                varnames = ['parameter_0']
            elif isinstance(varnames, list):
                varnames = varnames[0]
            return _Trace(_Hashmap(**dict([(varnames[0], chain)])))
        if varnames is None:
            varnames = ['parameter_{}'.format(i) for i in range(len(chain))]
        else:
            if len(varnames) != new_p:
                raise NotImplementedError('Parameter Subsetting by varnames '
                                          'is not currenlty implented for raw arrays')
        return _Trace(*[_Hashmap(**{k: run.T[p] for p, k in enumerate(varnames)})
                      for run in chain])



from pysal.lib import cg
import numpy as np


def compute_length(v0, v1):
    """Compute the euclidean distance between two points.
    
    Parameters
    ----------
    
    v0 : tuple
        sequence in the form x, y
    
    vq : tuple
        sequence in the form x, y
    
    Returns
    --------
    
    euc_dist : float
        Euclidean distance
    
    Examples
    --------
    
    >>> import pysal.explore.spaghetti as spgh
    >>> point1, point2 = (0,0), (1,1)
    >>> spgh.util.compute_length(point1, point2)
    1.4142135623730951
    """
    euc_dist = np.sqrt((v0[0] - v1[0])**2 + (v0[1] - v1[1])**2)
    return euc_dist


def get_neighbor_distances(ntw, v0, l):
    """Get distances to the nearest node neighbors along connecting edges.
    
    Parameters
    ----------
    ntw : spaghetti.Network
        spaghetti Network object.
    v0 : int
        Node id
    l : dict
        key is tuple (start node, end node); value is float.
        Cost per edge to travel, e.g. distance.
    
    Returns
    -------
    neighbors : dict
        key is int (node id); value is float (distance)
    
    Examples
    --------
    >>> import pysal.explore.spaghetti as spgh
    >>> from pysal.lib import examples
    >>> ntw = spgh.Network(examples.get_path('streets.shp'))
    >>> neighs = spgh.util.get_neighbor_distances(ntw, 0, ntw.edge_lengths)
    >>> neighs[1]
    102.62353453439829
    """
    edges = ntw.enum_links_node(v0)
    neighbors = {}
    for e in edges:
        if e[0] != v0:
            neighbors[e[0]] = l[e]
        else:
            neighbors[e[1]] = l[e]
    return neighbors


def generatetree(pred):
    """Rebuild the shortest path from root origin to destination
    
    Parameters
    ----------
    
    pred : list
        List of preceeding nodes for traversal route.
    
    Returns
    --------
    
    tree : dict
        key is root origin; value is root origin to destination.
    
    Examples
    --------
    
    >>> import pysal.explore.spaghetti as spgh
    >>> from pysal.lib import examples
    >>> ntw = spgh.Network(examples.get_path('streets.shp'))
    >>> distance, pred = spgh.util.dijkstra(ntw, ntw.edge_lengths, 0)
    >>> tree = spgh.util.generatetree(pred)
    >>> tree[3]
    [23, 22, 20, 19, 170, 2, 0]
    """
    tree = {}
    for i, p in enumerate(pred):
        if p == -1:
            # root node
            tree[i] = [i]
            continue
        idx = p
        path = [idx]
        while idx >= 0:
            nextnode = pred[idx]
            idx = nextnode
            if idx >= 0:
                path.append(nextnode)
        tree[i] = path
    return tree


def dijkstra(ntw, cost, v0, n=float('inf')):
    """Compute the shortest path between a start node and all other nodes in
    an origin-destination matrix.

    Parameters
    ----------
    
    ntw :  spaghetti.Network
        spaghetti Network object.
    
    cost : dict
        key is tuple (start node, end node); value is float.
        Cost per edge to travel, e.g. distance.
    
    v0 : int
        Start node ID
    
    n : float
        integer break point to stop iteration and return n neighbors.
        Default is ('inf').
    
    Returns
    -------
    
    distance : list
        List of distances from node to all other nodes.
    
    pred : list
        List of preceeding nodes for traversal route.
    
    Notes
    -----
    
    Based on :cite:`Dijkstra1959a`.
    
    Examples
    --------
    
    >>> import pysal.explore.spaghetti as spgh
    >>> from pysal.lib import examples
    >>> ntw = spgh.Network(examples.get_path('streets.shp'))
    >>> distance, pred = spgh.util.dijkstra(ntw, ntw.edge_lengths, 0)
    >>> round(distance[196], 4)
    5505.6682
    >>> pred[196]
    133
    """
    distance = [n for x in ntw.node_list]
    idx = ntw.node_list.index(v0)
    distance[ntw.node_list.index(v0)] = 0
    pred = [-1 for x in ntw.node_list]
    a = set()
    a.add(v0)
    while len(a) > 0:
        # Get node with the lowest value from distance.
        dist = n
        for node in a:
            if distance[node] < dist:
                dist = distance[node]
                v = node
        # Remove that node from the set.
        a.remove(v)
        last = v
        # 4. Get the neighbors to the current node.
        neighbors = get_neighbor_distances(ntw, v, cost)
        for v1, indiv_cost in neighbors.items():
            if distance[v1] > distance[v] + indiv_cost:
                distance[v1] = distance[v] + indiv_cost
                pred[v1] = v
                a.add(v1)
    pred = np.array(pred, dtype=np.int)
    return distance, pred


def dijkstra_mp(ntw_cost_node):
    """
    Compute the shortest path between a start node and all other
    nodes in the web utilizing multiple cores upon request.
    
    Parameters
    ----------
    
    ntw_cost_node : tuple
        tuple of arguments to pass into dijkstra
        (1) ntw - spaghetti.Network; spaghetti Network object; (2) cost - dict;
        key is tuple (start node, end node); value is float - Cost per edge to
        travel, e.g. distance; (3) node - int; Start node ID
    
    Returns
    -------
    
    distance : list
        List of distances from node to all other nodes.
    
    pred : list
        List of preceeding nodes for traversal route.
    
    Notes
    -----
    
    Based on :cite:`Dijkstra1959a`.
    
    Examples
    --------
    
    >>> import pysal.explore.spaghetti as spgh
    >>> from pysal.lib import examples
    >>> ntw = spgh.Network(examples.get_path('streets.shp'))
    >>> distance, pred = spgh.util.dijkstra(ntw, ntw.edge_lengths, 0)
    >>> round(distance[196], 4)
    5505.6682
    >>> pred[196]
    133
    """
    ntw, cost, node = ntw_cost_node
    distance, pred = dijkstra(ntw, cost, node)
    return distance, pred


def squared_distance_point_segment(point, segment):
    """Find the squared distance between a point and a segment.
    
    Parameters
    ----------
    
    point : tuple
        point coordinates (x,y)
    
    segment : list
        List of 2 point coordinate tuples [(x0,y0), (x1,y1)].
    
    Returns
    -------
    sqd : float
        distance squared between point and segment
    
    nearp : numpy.ndarray
        array of (xb, yb); the nearest point on the segment
    
    Examples
    --------
    
    >>> import pysal.explore.spaghetti as spgh
    >>> point, segment = (1,1), ((0,0), (2,0))
    >>> spgh.util.squared_distance_point_segment(point, segment)
    (1.0, array([1., 0.]))
    """
    #
    p0, p1 = [np.array(p) for p in segment]
    v = p1 - p0
    p = np.array(point)
    w = p - p0
    c1 = np.dot(w, v)
    if c1 <= 0.:
        sqd = np.dot(w.T, w)
        nearp = p0
        return sqd, nearp
    #
    c2 = np.dot(v, v)
    if c2 <= c1:
        dp1 = p - p1
        sqd = np.dot(dp1.T, dp1)
        nearp = p1
        return sqd, nearp
    #
    b = c1 / c2
    bv = np.dot(b, v)
    pb = p0 + bv
    d2 = p - pb
    sqd = np.dot(d2, d2)
    nearp = pb
    return sqd, nearp


def snap_points_on_segments(points, segments):
    """Place points onto closet segment in a set of segments
    
    Parameters
    ----------
    
    points : dict
        Point id as key and (x,y) coordinate as value
    
    segments : list
        Elements are of type pysal.lib.cg.shapes.Chain
        ** Note ** each element is a segment represented as a chain with
        *one head and one tail node* in other words one link only.
    
    Returns
    -------
    
    p2s : dict
        key [point id (see points in arguments)]; value [a 2-tuple 
        ((head, tail), point) where (head, tail) is the target segment,
        and point is the snapped location on the segment.
    
    Examples
    --------
    
    >>> import pysal.explore.spaghetti as spgh
    >>> from pysal.lib.cg.shapes import Point, Chain
    >>> points = {0: Point((1,1))}
    >>> segments = [Chain([Point((0,0)), Point((2,0))])]
    >>> spgh.util.snap_points_on_segments(points, segments)
    {0: ([(0.0, 0.0), (2.0, 0.0)], array([1., 0.]))}
    """
    
    # Put segments in an Rtree.
    rt = cg.Rtree()
    SMALL = np.finfo(float).eps
    node2segs = {}
    
    for segment in segments:
        head, tail = segment.vertices
        x0, y0 = head
        x1, y1 = tail
        if (x0, y0) not in node2segs:
            node2segs[(x0, y0)] = []
        if (x1, y1) not in node2segs:
            node2segs[(x1, y1)] = []
        node2segs[(x0, y0)].append(segment)
        node2segs[(x1, y1)].append(segment)
        x0, y0, x1, y1 = segment.bounding_box
        x0 -= SMALL
        y0 -= SMALL
        x1 += SMALL
        y1 += SMALL
        r = cg.Rect(x0, y0, x1, y1)
        rt.insert(segment, r)
        
    # Build a KDtree on segment nodes.
    kt = cg.KDTree(list(node2segs.keys()))
    p2s = {}
    
    for ptIdx, point in points.items():
        # First, find nearest neighbor segment node for the point.
        dmin, node = kt.query(point, k=1)
        node = tuple(kt.data[node])
        closest = node2segs[node][0].vertices
        
        # Use this segment as the candidate closest segment:  closest
        # Use the distance as the distance to beat:           dmin
        p2s[ptIdx] = (closest, np.array(node))
        x0 = point[0] - dmin
        y0 = point[1] - dmin
        x1 = point[0] + dmin
        y1 = point[1] + dmin
        
        # Find all segments with bounding boxes that intersect
        # a query rectangle centered on the point with sides of length 2*dmin.
        candidates = [cand for cand in rt.intersection([x0, y0, x1, y1])]
        dmin += SMALL
        dmin2 = dmin * dmin
        
        # Of the candidate segments, find the nearest to the query point.
        for candidate in candidates:
            dnc, p2b = squared_distance_point_segment(point,
                                                      candidate.vertices)
            if dnc <= dmin2:
                closest = candidate.vertices
                dmin2 = dnc
                p2s[ptIdx] = (closest, p2b)
                
    return p2s



from __future__ import division
import numpy as np
import copy
import scipy.linalg as scla
import scipy.sparse as spar
from scipy import stats
from scipy.spatial import distance as d
from .utils import explode, nexp
from .sample import sample_phi, logp_phi
from pysal.model.spvcm.abstracts import Sampler_Mixin, Trace, Hashmap
from pysal.model.spvcm.verify import center as verify_center, covariates as verify_covariates
from pysal.model.spvcm.utils import chol_mvn
from pysal.model.spvcm.steps import Metropolis, Slice

class SVC(Sampler_Mixin):
    """
    A class to initialize a spatially-varying coefficient model

    Parameters to be estimated:
    Beta    : effects at each point, distributed around Mu _\beta
    Mus     : hierarchical mean of Beta effects
    T       : global covariance matrix for effects
    Phi     : spatial dependence parameter for effects
    Tau2    : prediction error for the model

    The distance matrix will be divided by the maximum distance,
    so that it will vary between 0 and 1.

    Hyperparameters as given:
    a0          : tau location hyperparameter
    b0          : tau scale hyperparameter
    v0          : T matrix degrees of freedom parameter
    Omega       : T matrix covariance prior
    mu0         : Mean beta hyperparameter
    mu_cov0      : Beta hyperparameter for Covariance
    phi_shape0  : Phi spatial dependence shape hyperparameter
    phi_rate0   : Phi spatial dependence rate hyperparameter
    phi_scale0  : Phi spatial dependence scale hyperparameter, overrides the
                  rate parameter. always 1/phi_rate0
    """
    def __init__(self,
                 #data parameters
                 Y, X, coordinates, n_samples=1000, n_jobs=1,
                 priors=None,
                 configs=None,
                 starting_values=None,
                 extra_traced_params = None,
                 dmetric='euclidean',
                 correlation_function=nexp,
                 verbose=False,
                 center=True,
                 rescale_dists=True):
        if center:
            X = verify_center(X)
        X = verify_covariates(X)

        N,p = X.shape
        Xs = X

        X = explode(X)
        self.state = Hashmap(X=X, Y=Y, coordinates=coordinates)
        self.traced_params = ['Betas', 'Mus', 'T', 'Phi', 'Tau2']
        if extra_traced_params is not None:
            self.traced_params.extend(extra_traced_params)
        self.trace = Trace(**{param:[] for param in self.traced_params})
        st = self.state
        self.state.correlation_function = correlation_function
        self.verbose = verbose


        st.Y = Y
        st.X = X
        st.Xs = Xs
        st.N = N
        st.p = p

        st._dmetric = dmetric
        if isinstance(st._dmetric, str):
            st.pwds = d.squareform(d.pdist(st.coordinates, metric=st._dmetric))
        elif callable(st._dmetric):
            st.pwds = st._dmetric(st.coordinates)

        st.max_dist = st.pwds.max()
        if rescale_dists:
            st.pwds = st.pwds/st.max_dist
            st._old_max = st.max_dist
            st.max_dist = 1.

        if configs is None:
            configs = dict()
        if priors is None:
            priors = dict()
        if starting_values is None:
            starting_values = dict()

        self._setup_priors(**priors)
        self._setup_starting_values(**starting_values)
        self._setup_configs(**configs)

        self._verbose = verbose
        self.cycles = 0

        if n_samples > 0:
            try:
                self.sample(n_samples, n_jobs=n_jobs)
            except (np.linalg.LinAlgError, ValueError) as e:
                Warn('Encountered the following LinAlgError. '
                     'Model will return for debugging. \n {}'.format(e))

    def _setup_priors(self, Tau2_a0=.001, Tau2_b0=.001,
                            T_v0 = 3, T_Omega0=None,
                            Mus_mean0 = None, Mus_cov0=None,
                            Phi_shape0=1, Phi_rate0=None):
        st = self.state
        st.Tau2_a0 = Tau2_a0
        st.Tau2_b0 = Tau2_b0
        st.T_v0 = T_v0
        if T_Omega0 is None:
            st.Ip = np.identity(st.p)
            T_Omega0 = .1 * st.Ip
        st.T_Omega0 = T_Omega0
        if type(Mus_mean0) in (float, int):
            Mus_mean0 = np.ones((st.p,1)) * Mus_mean0
        elif Mus_mean0 is None:
            Mus_mean0 = np.zeros((st.p, 1))
        st.Mus_mean0 = Mus_mean0
        if Mus_cov0 is None:
            Mus_cov0 = 1000*st.Ip
        st.Mus_cov0 = Mus_cov0
        st.Phi_shape0 = Phi_shape0
        if Phi_rate0 is None:
           Phi_rate0 = st.Phi_shape0 / ((-.5*st.pwds.max() / np.log(.05)))
        st.Phi_rate0 = Phi_rate0

    def _setup_starting_values(self, Phi=None, T=None,
                                     Mus=None, Betas=None, Tau2=2):
        """
        Setup initial values of the sampler for parameters

        Defaults
        ---------
        Phi     : 3*shape/rate for the parameter by default.
        T       : an identity matrix
        Mu      : zeros
        Beta    : zeros
        """
        if T is None:
            T = np.eye(self.state.p)
        self.state.T = T
        if Mus is None:
            Mus = np.zeros((1,self.state.p))
        self.state.Mus = Mus
        if Betas is None:
            Betas = np.zeros((self.state.p * self.state.N, 1))
        self.state.Betas = Betas
        if Phi is None:
            Phi = 3*self.state.Phi_shape0 / self.state.Phi_rate0
        self.state.Phi = Phi
        self.state.Tau2=Tau2

    def _setup_configs(self, Phi_method = 'met', Phi_configs=None, **uncaught):
        """
        Setup Configs of the MCMC sampled parameters in the model.
        """
        if Phi_method.lower().startswith('met'):
            method = Metropolis
        elif Phi_method.lower().startswith('slice'):
            method = Slice
        else:
            raise Exception('`Phi_method` option not understood. `{}` provided'             .format(Phi_method))

        if uncaught != dict() and Phi_configs is None:
            Phi_configs = uncaught
        elif Phi_configs is not None and uncaught == dict():
            if isinstance(Phi_configs, dict):
                Phi_configs = Phi_configs
            else:
                raise TypeError('Type of `Phi_configs` not understood. Must be'
                                'dict containing the configuration for the MCMC step sampling Phi. Recieved object of type:'
                                '\n{}'.format(type(Phi_configs)))
        elif Phi_configs is None and uncaught == dict():
            Phi_configs = uncaught
        else:
            raise Exception('Uncaught options {} passed in addition to '
                            '`Phi_configs` {}.'.format(uncaught, Phi_configs))

        self.configs = Hashmap()
        self.configs.Phi = method('Phi', logp_phi, **Phi_configs)


    def _finalize(self):
        """
        Compute derived quantities that make the sampling loop more efficient.
        Once computed, the priors are considered "set."
        """
        st = self.state
        st = self.state
        st.In = np.identity(st.N)
        st.XtX = st.X.T.dot(st.X)
        st.iota_n = np.ones((st.N,1))
        st.Xty = st.X.T.dot(st.Y)
        st.np2n = np.zeros((st.N * st.p, st.N))
        for i in range(st.N):
            st.np2n[i*st.p:(i+1)*st.p, i] = 1
        st.np2p = np.vstack([np.eye(st.p) for _ in range(st.N)])
        st.Mus_cov0i = scla.inv(st.Mus_cov0)
        st.Mus_kernel_prior = np.dot(st.Mus_cov0i, st.Mus_mean0)
        st.Tau2_an = st.Tau2_a0 + st.N/2.
        st.T_vn = st.T_v0 + st.N

    def _fuzz_starting_values(self):
        super(SVC, self)._fuzz_starting_values()
        self.state.Phi += np.random.uniform(0,10)
        self.state.Mus += np.random.normal(0,10,size=self.state.Mus.shape)
        self.state.Tau2 += np.random.uniform(0,10)

    def _iteration(self):
        """
        Conduct one iteration of a Gibbs sampler for the self using the state
        provided.
        """
        st = self.state

        ## Tau, EQ 3 in appendix of Wheeler & Calder
        ## Inverse Gamma w/ update to scale, no change to dof
        y_Xbeta = st.Y - st.X.dot(st.Betas)
        st.Tau2_bn = st.Tau2_b0 + .5 * y_Xbeta.T.dot(y_Xbeta)
        st.Tau2 = stats.invgamma.rvs(st.Tau2_an, scale=st.Tau2_bn)

        ##covariance: T, EQ 4 in appendix of Wheeler & Calder
        ## inverse wishart w/ update to covariance matrix, no change to dof
        st.H = st.correlation_function(st.Phi, st.pwds)
        st.Hinv = scla.inv(st.H)
        st.tiled_Hinv = np.linalg.multi_dot([st.np2n, st.Hinv, st.np2n.T])
        st.tiled_Mus = np.kron(st.iota_n, st.Mus.reshape(-1,1))
        st.info = (st.Betas - st.tiled_Mus).dot((st.Betas - st.tiled_Mus).T)
        st.kernel = np.multiply(st.tiled_Hinv, st.info)
        st.covm_update = np.linalg.multi_dot([st.np2p.T, st.kernel, st.np2p])
        st.T = stats.invwishart.rvs(df=st.T_vn, scale=(st.covm_update + st.T_Omega0))

        ##mean hierarchical effects: mu_\beta, in EQ 5 of Wheeler & Calder
        ##normal with both a scale and a location update, priors don't change
        #compute scale of mu_\betas
        st.Sigma_beta = np.kron(st.H, st.T)
        st.Psi = np.linalg.multi_dot((st.X, st.Sigma_beta, st.X.T)) + st.Tau2 * st.In
        Psi_inv = scla.inv(st.Psi)
        S_notinv_update = np.linalg.multi_dot((st.Xs.T, Psi_inv, st.Xs))
        S = scla.inv(st.Mus_cov0i + S_notinv_update)

        #compute location of mu_\betas
        mkernel_update = np.linalg.multi_dot((st.Xs.T, Psi_inv, st.Y))
        st.Mus_mean = np.dot(S, (mkernel_update + st.Mus_kernel_prior))

        #draw them using cholesky decomposition: N(m, Sigma) = m + chol(Sigma).N(0,1)
        st.Mus = chol_mvn(st.Mus_mean, S)
        st.tiled_Mus = np.kron(st.iota_n, st.Mus)

        ##effects \beta, in equation 6 of Wheeler & Calder
        ##Normal with an update to both scale and location, priors don't change

        #compute scale of betas
        st.Tinv = scla.inv(st.T)
        st.kronHiTi = np.kron(st.Hinv, st.Tinv)
        Ai = st.XtX / st.Tau2 + st.kronHiTi
        A = scla.inv(Ai)

        #compute means of betas
        C = st.Xty / st.Tau2 + np.dot(st.kronHiTi, st.tiled_Mus)
        st.Betas_means = np.dot(A, C)
        st.Betas_cov = A

        #draw them using cholesky decomposition
        st.Betas = chol_mvn(st.Betas_means, st.Betas_cov)

        # local nonstationarity parameter Phi, in equation 7 in Wheeler & Calder
        # sample using metropolis
        st.Phi = self.configs.Phi(st)



import numpy as np
import matplotlib.pyplot as plt

"""
Lightweight visualizations for pysal.lib using Matplotlib and Geopandas

TODO
* make gdf argument in plot_spatial_weights optional
"""

__author__ = ("Stefanie Lumnitz <stefanie.lumitz@gmail.com>")



def plot_spatial_weights(w, gdf, indexed_on=None, ax=None, 
                         figsize=(10,10), node_kws=None, edge_kws=None,
                         nonplanar_edge_kws=None):
    """
    Plot spatial weights network.
    NOTE: Additionally plots `w.non_planar_joins` if
    `pysal.lib.weights.util.nonplanar_neighbors()` was applied.

    Arguments
    ---------
    w : pysal.lib.W object
        Values of pysal.lib weights object.
    gdf : geopandas dataframe 
        The original shapes whose topological relations are 
        modelled in W.
    indexed_on : str, optional
        Column of gdf which the weights object uses as an index.
        Default =None, so the geodataframe's index is used.
    ax : matplotlib axis, optional
        Axis on which to plot the weights. 
        Default =None, so plots on the current figure.
    figsize : tuple, optional
        W, h of figure. Default =(10,10)
    node_kws : keyword argument dictionary, optional
        Dictionary of keyword arguments to send to pyplot.scatter,
        which provide fine-grained control over the aesthetics
        of the nodes in the plot. Default =None.
    edge_kws : keyword argument dictionary, optional
        Dictionary of keyword arguments to send to pyplot.plot,
        which provide fine-grained control over the aesthetics
        of the edges in the plot. Default =None.
    nonplanar_edge_kws : keyword argument dictionary, optional
        Dictionary of keyword arguments to send to pyplot.plot,
        which provide fine-grained control over the aesthetics
        of the edges from `weights.non_planar_joins` in the plot.
        Default =None.

    Returns
    -------
    fig : matplotlip Figure instance
        Figure of spatial weight network.
    ax : matplotlib Axes instance
        Axes in which the figure is plotted. 

    Examples
    --------
    Imports
    
    >>> from pysal.lib.weights.contiguity import Queen
    >>> import geopandas as gpd
    >>> import pysal.lib
    >>> from pysal.lib import examples
    >>> import matplotlib.pyplot as plt
    >>> from pysal.viz.splot.libpysal import plot_spatial_weights
    
    Data preparation and statistical analysis
    
    >>> gdf = gpd.read_file(examples.get_path('map_RS_BR.shp'))
    >>> weights = Queen.from_dataframe(gdf)
    >>> wnp = pysal.lib.weights.util.nonplanar_neighbors(weights, gdf)
    
    Plot weights
    
    >>> plot_spatial_weights(weights, gdf)
    >>> plt.show()
    
    Plot corrected weights
    
    >>> plot_spatial_weights(wnp, gdf)
    >>> plt.show()
    
    """
    if ax is None:
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111)
    else:
        fig = ax.get_figure()
    
    # default for node_kws
    if node_kws is None:
        node_kws = dict(marker='.', s=10, color='#4d4d4d')
    
    # default for edge_kws
    if edge_kws is None:
        edge_kws = dict(color='#4393c3')
    
    # default for nonplanar_edge_kws
    if nonplanar_edge_kws is None:
        edge_kws.setdefault('lw', 0.7)
        nonplanar_edge_kws = edge_kws.copy()
        nonplanar_edge_kws['color'] = '#d6604d'

    node_has_nonplanar_join = []
    if hasattr(w, 'non_planar_joins'):
        # This attribute is present when an instance is created by the user
        # calling `weights.util.nonplanar_neighbors`. If so, treat those
        # edges differently by default.
        node_has_nonplanar_join = w.non_planar_joins.keys()

    # Plot the polygons from the geodataframe as a base layer
    gdf.plot(ax=ax, color='#bababa', edgecolor='w')

    for idx, neighbors in w:
        if idx in w.islands:
            continue

        if indexed_on is not None:
            neighbors = gdf[gdf[indexed_on].isin(neighbors)].index.tolist()
            idx = gdf[gdf[indexed_on] == idx].index.tolist()[0]

        # Find centroid of each polygon that's a neighbor, as a numpy array (x, y)
        centroids = gdf.loc[neighbors].centroid.apply(lambda p: (p.x, p.y))
        centroids = np.vstack(centroids.values)
        # Find the centroid of the polygon we're looking at now
        focal = np.hstack(gdf.loc[idx].geometry.centroid.xy)
        seen = set()
        for nidx, neighbor in zip(neighbors, centroids):
            if (idx, nidx) in seen:
                continue
            seen.update((idx, nidx))
            seen.update((nidx, idx))

            # Plot current edge as line: plot([x0, x1], [y0, y1])
            if (idx in node_has_nonplanar_join) and (nidx in w.non_planar_joins[idx]):
                # This is a non-planar edge
                ax.plot(*list(zip(focal, neighbor)), marker=None, **nonplanar_edge_kws)
            else:
                ax.plot(*list(zip(focal, neighbor)), marker=None, **edge_kws)

    # Plot the nodes
    ax.scatter(gdf.centroid.apply(lambda p: p.x),
               gdf.centroid.apply(lambda p: p.y),
               **node_kws)
    ax.set_axis_off()
    ax.set_aspect('equal')
    return fig, ax



from .generic import Generic
from .none import MVCM
from .se_se import SESE
from .se_sma import SESMA
from .sma_se import SMASE
from .sma_sma import SMASMA

try:
    del (generic, none, se_se, se_sma, sma_se, sma_sma)
except NameError:
    pass



import unittest
import numpy as np
from pysal.lib import cg, examples
from .. import util
from .. import network
try:
    import geopandas
    GEOPANDAS_EXTINCT = False
except ImportError:
    GEOPANDAS_EXTINCT = True


@unittest.skipIf(GEOPANDAS_EXTINCT, 'Missing Geopandas')
class TestNetwork(unittest.TestCase):
    
    def setUp(self):
        path_to_shp = examples.get_path('streets.shp')
        gdf = geopandas.read_file(path_to_shp)
        self.ntw = network.Network(in_data=gdf)
    
    def tearDown(self):
        pass
    
    def test_extract_network(self):
        self.assertEqual(len(self.ntw.edges), 303)
        self.assertEqual(len(self.ntw.nodes), 230)
        edgelengths = self.ntw.edge_lengths.values()
        self.assertAlmostEqual(sum(edgelengths), 104414.0920159, places=5)
        self.assertIn(0, self.ntw.adjacencylist[1])
        self.assertIn(0, self.ntw.adjacencylist[2])
        self.assertNotIn(0, self.ntw.adjacencylist[3])
    
    def test_contiguity_weights(self):
        w = self.ntw.contiguityweights(graph=False)
        self.assertEqual(w.n, 303)
        self.assertEqual(w.histogram,
                         [(2, 35), (3, 89), (4, 105), (5, 61), (6, 13)])
    
    def test_contiguity_weights_graph(self):
        w = self.ntw.contiguityweights(graph=True)
        self.assertEqual(w.n, 179)
        self.assertEqual(w.histogram,
                         [(2, 2), (3, 2), (4, 45), (5, 82), (6, 48)])
    
    def test_distance_band_weights(self):
        # I do not trust this result, test should be manually checked.
        w = self.ntw.distancebandweights(threshold=500)
        self.assertEqual(w.n, 230)
        self.assertEqual(w.histogram,
                         [(1, 22), (2, 58), (3, 63), (4, 40),
                          (5, 36), (6, 3), (7, 5), (8, 3)])
    
    def test_edge_segmentation(self):
        n200 = self.ntw.segment_edges(200.0)
        self.assertEqual(len(n200.edges), 688)
        n200 = None
    
    def test_enum_links_node(self):
        coincident = self.ntw.enum_links_node(24)
        self.assertIn((24, 48), coincident)


@unittest.skipIf(GEOPANDAS_EXTINCT, 'Missing Geopandas')
class TestNetworkPointPattern(unittest.TestCase):
    
    def setUp(self):
        path_to_shp = examples.get_path('streets.shp')
        gdf = geopandas.read_file(path_to_shp)
        self.ntw = network.Network(in_data=gdf)
        for obs in ['schools', 'crimes']:
            path_to_shp = examples.get_path('{}.shp'.format(obs))
            in_data = geopandas.read_file(path_to_shp)
            self.ntw.snapobservations(in_data, obs, attribute=True)
            setattr(self, obs, self.ntw.pointpatterns[obs])
    
    def tearDown(self):
        pass
    
    def test_add_point_pattern(self):
        self.assertEqual(self.crimes.npoints, 287)
        self.assertIn('properties', self.crimes.points[0])
        self.assertIn([1, 1], self.crimes.points[0]['properties'])
    
    def test_count_per_edge(self):
        counts = self.ntw.count_per_edge(
            self.ntw.pointpatterns['crimes'].obs_to_edge, graph=False)
        meancounts = sum(counts.values()) / float(len(counts.keys()))
        self.assertAlmostEqual(meancounts, 2.682242, places=5)
    
    def test_count_per_graph_edge(self):
        counts = self.ntw.count_per_edge(
            self.ntw.pointpatterns['crimes'].obs_to_edge, graph=True)
        meancounts = sum(counts.values()) / float(len(counts.keys()))
        self.assertAlmostEqual(meancounts, 3.29885, places=5)
    
    def test_simulate_normal_observations(self):
        npoints = self.ntw.pointpatterns['crimes'].npoints
        sim = self.ntw.simulate_observations(npoints)
        self.assertEqual(npoints, sim.npoints)
    
    def test_simulate_poisson_observations(self):
        npoints = self.ntw.pointpatterns['crimes'].npoints
        sim = self.ntw.simulate_observations(npoints, distribution='poisson')
        self.assertEqual(npoints, sim.npoints)
    
    def test_all_neighbor_distances(self):
        matrix1, tree = self.ntw.allneighbordistances('schools', gen_tree=True)
        known_mtx_val = 17682.436988
        known_tree_val = (173, 64)
        
        self.assertAlmostEqual(np.nansum(matrix1[0]), known_mtx_val, places=4)
        self.assertEqual(tree[(6, 7)], known_tree_val)
        
        for k, (distances, predlist) in self.ntw.alldistances.items():
            self.assertEqual(distances[k], 0)
            for p, plists in predlist.items():
                self.assertEqual(plists[-1], k)
                self.assertEqual(self.ntw.node_list, list(predlist.keys()))
                
        matrix2 = self.ntw.allneighbordistances('schools', fill_diagonal=0.)
        observed = matrix2.diagonal()
        known = np.zeros(matrix2.shape[0])
        self.assertEqual(observed.all(), known.all())
        
        matrix3 = self.ntw.allneighbordistances('schools', snap_dist=True)
        known_mtx_val = 3218.2597894
        observed_mtx_val = matrix3
        self.assertAlmostEqual(observed_mtx_val[0, 1], known_mtx_val, places=4)
    
    def test_nearest_neighbor_distances(self):
        # general test
        with self.assertRaises(KeyError):
            self.ntw.nearestneighbordistances('i_should_not_exist')
        nnd1 = self.ntw.nearestneighbordistances('schools')
        nnd2 = self.ntw.nearestneighbordistances('schools',
                                                 destpattern='schools')
        nndv1 = np.array(list(nnd1.values()))[:,1].astype(float)
        nndv2 = np.array(list(nnd2.values()))[:,1].astype(float)
        np.testing.assert_array_almost_equal_nulp(nndv1, nndv2)
        
        # nearest neighbor keeping zero test
        known_zero = ([19], 0.0)[0]
        nn_c = self.ntw.nearestneighbordistances('crimes',
                                                 keep_zero_dist=True)
        self.assertEqual(nn_c[18][0], known_zero)
        
        # nearest neighbor omitting zero test
        known_nonzero = ([11], 165.33982412719126)[1]
        nn_c = self.ntw.nearestneighbordistances('crimes',
                                                 keep_zero_dist=False)
        self.assertAlmostEqual(nn_c[18][1], known_nonzero, places=4)
        
        # nearest neighbor with snap distance
        known_neigh = ([3], 402.5219673922477)[1]
        nn_c = self.ntw.nearestneighbordistances('crimes',
                                                 keep_zero_dist=True,
                                                 snap_dist=True)
        self.assertAlmostEqual(nn_c[0][1], known_neigh, places=4)

@unittest.skipIf(GEOPANDAS_EXTINCT, 'Missing Geopandas')
class TestNetworkAnalysis(unittest.TestCase):
    
    def setUp(self):
        path_to_shp = examples.get_path('streets.shp')
        gdf = geopandas.read_file(path_to_shp)
        self.ntw = network.Network(in_data=gdf)
        pt_str = 'crimes'
        path_to_shp = examples.get_path('{}.shp'.format(pt_str))
        in_data = geopandas.read_file(path_to_shp)
        self.ntw.snapobservations(in_data, pt_str, attribute=True)
        npts = self.ntw.pointpatterns['crimes'].npoints
        self.ntw.simulate_observations(npts)
    
    def tearDown(self):
        pass
    
    def test_network_f(self):
        obtained = self.ntw.NetworkF(self.ntw.pointpatterns['crimes'],
                                     permutations=5, nsteps=20)
        self.assertEqual(obtained.lowerenvelope.shape[0], 20)
    
    def test_network_g(self):
        obtained = self.ntw.NetworkG(self.ntw.pointpatterns['crimes'],
                                     permutations=5, nsteps=20)
        self.assertEqual(obtained.lowerenvelope.shape[0], 20)
    
    def test_network_k(self):
        obtained = self.ntw.NetworkK(self.ntw.pointpatterns['crimes'],
                                     permutations=5, nsteps=20)
        self.assertEqual(obtained.lowerenvelope.shape[0], 20)


@unittest.skipIf(GEOPANDAS_EXTINCT, 'Missing Geopandas')
class TestNetworkUtils(unittest.TestCase):
    
    def setUp(self):
        path_to_shp = examples.get_path('streets.shp')
        gdf = geopandas.read_file(path_to_shp)
        self.ntw = network.Network(in_data=gdf)
    
    def tearDown(self):
        pass
    
    def test_compute_length(self):
        self.point1, self.point2 = (0,0), (1,1)
        self.length = util.compute_length( self.point1, self.point2)
        self.assertAlmostEqual(self.length, 1.4142135623730951, places=4)
    
    def test_get_neighbor_distances(self):
        self.known_neighs = {1: 102.62353453439829, 2: 660.000001049743}
        self.neighs = util.get_neighbor_distances(self.ntw, 0,
                                                  self.ntw.edge_lengths)
        self.assertAlmostEqual(self.neighs[1], 102.62353453439829, places=4)
        self.assertAlmostEqual(self.neighs[2], 660.000001049743, places=4)
    
    def test_generate_tree(self):
        self.known_path = [23, 22, 20, 19, 170, 2, 0]
        self.distance, self.pred = util.dijkstra(self.ntw,
                                                 self.ntw.edge_lengths, 0)
        self.tree = util.generatetree(self.pred)
        self.assertEqual(self.tree[3], self.known_path)
    
    def test_dijkstra(self):
        self.distance, self.pred = util.dijkstra(self.ntw,
                                                 self.ntw.edge_lengths, 0)
        self.assertAlmostEqual(self.distance[196], 5505.668247, places=4)
        self.assertEqual(self.pred[196], 133)
    
    def test_dijkstra_mp(self):
        self.distance, self.pred = util.dijkstra_mp((self.ntw,
                                                     self.ntw.edge_lengths, 0))
        self.assertAlmostEqual(self.distance[196], 5505.668247, places=4)
        self.assertEqual(self.pred[196], 133)
    
    def test_square_distance_point_segment(self):
        self.point, self.segment = (1,1), ((0,0), (2,0))
        self.sqrd_nearp = util.squared_distance_point_segment(self.point,
                                                              self.segment)
        self.assertEqual(self.sqrd_nearp[0], 1.0)
        self.assertEqual(self.sqrd_nearp[1].all(), np.array([1., 0.]).all())
    
    def test_snap_points_on_segments(self):
        self.points = {0: cg.shapes.Point((1,1))}
        self.segments = [cg.shapes.Chain([cg.shapes.Point((0,0)),
                                          cg.shapes.Point((2,0))])]
        self.snapped = util.snap_points_on_segments(self.points, self.segments)
        self.known_coords = [xy._Point__loc for xy in self.snapped[0][0]]
        self.assertEqual(self.known_coords, [(0.0, 0.0), (2.0, 0.0)])
        self.assertEqual(self.snapped[0][1].all(), np.array([1., 0.]).all())


if __name__ == '__main__':
    unittest.main()



#Bandwidth optimization methods

__author__ = "Taylor Oshan"

import numpy as np
from scipy import linalg
from copy import deepcopy
import copy
from collections import namedtuple

def golden_section(a, c, delta, function, tol, max_iter, int_score=False):
    """
    Golden section search routine
    Method: p212, 9.6.4
    Fotheringham, A. S., Brunsdon, C., & Charlton, M. (2002).
    Geographically weighted regression: the analysis of spatially varying relationships.

    Parameters
    ----------
    a               : float
                      initial max search section value
    b               : float
                      initial min search section value
    delta           : float
                      constant used to determine width of search sections
    function        : function
                      obejective function to be evaluated at different section
                      values
    int_score       : boolean
                      False for float score, True for integer score
    tol             : float
                      tolerance used to determine convergence
    max_iter        : integer
                      maximum iterations if no convergence to tolerance

    Returns
    -------
    opt_val         : float
                      optimal value
    opt_score       : kernel
                      optimal score
    output          : list of tuples
                      searching history
    """
    b = a + delta * np.abs(c-a)
    d = c - delta * np.abs(c-a)
    score = 0.0
    diff = 1.0e9
    iters  = 0
    output = []
    dict = {}
    while np.abs(diff) > tol and iters < max_iter:
        iters += 1
        if int_score:
            b = np.round(b)
            d = np.round(d)
        
        if b in dict:
            score_b = dict[b]
        else:
            score_b = function(b)
            dict[b] = score_b
        
        if d in dict:
            score_d = dict[d]
        else:
            score_d = function(d)
            dict[d] = score_d

        if score_b <= score_d:
            opt_val = b
            opt_score = score_b
            c = d
            d = b
            b = a + delta * np.abs(c-a)
            #if int_score:
                #b = np.round(b)
        else:
            opt_val = d
            opt_score = score_d
            a = b
            b = d
            d = c - delta * np.abs(c-a)
            #if int_score:
                #d = np.round(b)

        #if int_score:
        # opt_val = np.round(opt_val)
        output.append((opt_val, opt_score))
        diff = score_b - score_d
        score = opt_score
    return np.round(opt_val, 2), opt_score, output

def equal_interval(l_bound, u_bound, interval, function, int_score=False):
    """
    Interval search, using interval as stepsize

    Parameters
    ----------
    l_bound         : float
                      initial min search section value
    u_bound         : float
                      initial max search section value
    interval        : float
                      constant used to determine width of search sections
    function        : function
                      obejective function to be evaluated at different section
                      values
    int_score       : boolean
                      False for float score, True for integer score

    Returns
    -------
    opt_val         : float
                      optimal value
    opt_score       : kernel
                      optimal score
    output          : list of tuples
                      searching history
    """
    a = l_bound
    c = u_bound
    b = a + interval
    if int_score:
        a = np.round(a,0)
        c = np.round(c,0)
        b = np.round(b,0)

    output = []

    score_a = function(a)
    score_c = function(c)

    output.append((a,score_a))
    output.append((c,score_c))

    if score_a < score_c:
        opt_val = a
        opt_score = score_a
    else:
        opt_val = c
        opt_score = score_c

    while b < c:
        score_b = function(b)

        output.append((b,score_b))

        if score_b < opt_score:
            opt_val = b
            opt_score = score_b
        b = b + interval

    return opt_val, opt_score, output

def multi_bw(init, y, X, n, k, family, tol, max_iter, rss_score,
        gwr_func, bw_func, sel_func, multi_bw_min, multi_bw_max):
    """
    Multiscale GWR bandwidth search procedure using iterative GAM backfitting
    """
    if init is None:
        bw = sel_func(bw_func(y, X))
        optim_model = gwr_func(y, X, bw)
    else:
        optim_model = gwr_func(y, X, init)
     
    S = optim_model.S
    err = optim_model.resid_response.reshape((-1,1))
    param = optim_model.params
    
    R = np.zeros((n,n,k))
    
    for j in range(k):
        for i in range(n):
            wi = optim_model.W[i].reshape(-1,1)
            xT = (X * wi).T
            P = linalg.solve(xT.dot(X), xT)
            R[i,:,j] = X[i,j]*P[j]

    XB = np.multiply(param, X)
    if rss_score:
        rss = np.sum((err)**2)
    iters = 0
    scores = []
    delta = 1e6
    BWs = []
    VALs = []
    FUNCs = []
    
    try:
        from tqdm import tqdm #if they have it, let users have a progress bar
    except ImportError:
        def tqdm(x): #otherwise, just passthrough the range
            return x
    for iters in tqdm(range(1, max_iter+1)):
        new_XB = np.zeros_like(X)
        bws = []
        vals = []
        funcs = []
        current_partial_residuals = []
        params = np.zeros_like(X)
        f_XB = XB.copy()
        f_err = err.copy()
        
        for j in range(k):
            temp_y = XB[:,j].reshape((-1,1))
            temp_y = temp_y + err
            temp_X = X[:,j].reshape((-1,1))
            bw_class = bw_func(temp_y, temp_X)
            funcs.append(bw_class._functions)
            bw = sel_func(bw_class, multi_bw_min[j], multi_bw_max[j])
            optim_model = gwr_func(temp_y, temp_X, bw)
            Aj = optim_model.S
            new_Rj = Aj - np.dot(Aj, S) + np.dot(Aj, R[:,:,j])
            S = S - R[:,:,j] + new_Rj
            R[:,:,j] = new_Rj
            
            err = optim_model.resid_response.reshape((-1,1))
            param = optim_model.params.reshape((-1,))

            new_XB[:,j] = optim_model.predy.reshape(-1)
            bws.append(copy.deepcopy(bw))
            params[:,j] = param
            vals.append(bw_class.bw[1])
            current_partial_residuals.append(err.copy())

        num = np.sum((new_XB - XB)**2)/n
        den = np.sum(np.sum(new_XB, axis=1)**2)
        score = (num/den)**0.5
        XB = new_XB

        if rss_score:
            predy = np.sum(np.multiply(params, X), axis=1).reshape((-1,1))
            new_rss = np.sum((y - predy)**2)
            score = np.abs((new_rss - rss)/new_rss)
            rss = new_rss
        scores.append(copy.deepcopy(score))
        delta = score
        BWs.append(copy.deepcopy(bws))
        VALs.append(copy.deepcopy(vals))
        FUNCs.append(copy.deepcopy(funcs))
        if delta < tol:
            break

    opt_bws = BWs[-1]
    return (opt_bws, np.array(BWs),
                          np.array(scores), params,
                          err, S, R)



"""
Voronoi tesslation of 2-d point sets


Adapted from https://gist.github.com/pv/8036995

"""
import numpy as np
from scipy.spatial import Voronoi


__author__ = "Serge Rey <sjsrey@gmail.com>"

__all__ = ['voronoi_frames']

def voronoi(points, radius=None):
    """
    Determine finite Voronoi diagram for a 2-d point set 


    Parameters
    ----------
    points      : array-like
                  nx2 array of points

    radius      : float (optional) 
                  distance to 'points at infinity'

    Returns
    -------

    regions    : list
                  each element of the list contains sequence of the indexes of Voronoi vertices composing a Voronoi polygon (region)

    coordinates : array
                  coordinates of the Voronoi vertices


    Examples
    --------
    >>> points = [(10.2, 5.1), (4.7, 2.2), (5.3, 5.7), (2.7, 5.3)]
    >>> regions, coordinates = voronoi(points)
    >>> regions
    [[1, 3, 2], [4, 5, 1, 0], [0, 1, 7, 6], [9, 0, 8]]
    >>> coordinates
    array([[  4.21783296,   4.08408578],
           [  7.51956025,   3.51807539],
           [  9.4642193 ,  19.3994576 ],
           [ 14.98210684, -10.63503022],
           [ -9.22691341,  -4.58994414],
           [ 14.98210684, -10.63503022],
           [  1.78491801,  19.89803294],
           [  9.4642193 ,  19.3994576 ],
           [  1.78491801,  19.89803294],
           [ -9.22691341,  -4.58994414]])
    """
    vor = Voronoi(points)
    return voronoi_regions(vor, radius=radius)

def voronoi_regions(vor, radius=None):
    """
    Finite voronoi regions for a 2-d point set.


    Parameters
    ----------

    vor:  Voronoi (scipy.spatial)

    radius: float (optional)
            Distance to 'points at infinity'
    """
    new_regions = []
    new_vertices = vor.vertices.tolist()

    center = vor.points.mean(axis=0)
    if radius is None:
        radius = vor.points.ptp().max()*2

    all_ridges = {}
    for (p1, p2), (v1, v2) in zip(vor.ridge_points, vor.ridge_vertices):
        all_ridges.setdefault(p1, []).append((p2, v1, v2))
        all_ridges.setdefault(p2, []).append((p1, v1, v2))

    for p1, region in enumerate(vor.point_region):
        vertices = vor.regions[region]

        if all(v >= 0 for v in vertices):
            new_regions.append(vertices)
            continue

        ridges = all_ridges[p1]
        new_region = [v for v in vertices if v >= 0]

        for p2, v1, v2 in ridges:
            if v2 < 0:
                v1, v2 = v2, v1
            if v1 >= 0:
                continue

            t = vor.points[p2] - vor.points[p1] 
            t /= np.linalg.norm(t)
            n = np.array([-t[1], t[0]])

            midpoint = vor.points[[p1, p2]].mean(axis=0)
            direction = np.sign(np.dot(midpoint - center, n)) * n
            far_point = vor.vertices[v2] + direction * radius

            new_region.append(len(new_vertices))
            new_vertices.append(far_point.tolist())

        vs = np.asarray([new_vertices[v] for v in new_region])
        c = vs.mean(axis=0)
        angles = np.arctan2(vs[:, 1] - c[1], vs[:,0] - c[0])
        new_region = np.array(new_region)[np.argsort(angles)]

        new_regions.append(new_region.tolist())

    return new_regions, np.asarray(new_vertices)


def as_dataframes(regions, vertices, points):
    """
    Helper function to store finite Voronoi regions and originator points as
    geopandas (or pandas) data frames


    Parameters
    ----------

    regions     : list
                  each element of the list contains sequence of the indexes of
                  voronoi vertices composing a vornoi polygon (region)

    vertices    : array
                  coordinates of the vornoi vertices

    points      : array-like
                  originator points


    Returns
    -------

    region_df   : GeoDataFrame
                  Finite Voronoi polygons as geometries

    points_df   : GeoDataFrame
                  Originator points as geometries
    """
    try:
        import geopandas as gpd
    except ImportError:
        gpd = None

    try:
        from shapely.geometry import Polygon, Point
    except ImportError:
        from .shapes import Polygon, Point

    if gpd is not None:
        region_df = gpd.GeoDataFrame()
        region_df['geometry'] = [Polygon(vertices[region]) for region in regions]

        point_df = gpd.GeoDataFrame()
        point_df['geometry'] = gpd.GeoSeries(Point(pnt) for pnt in points)
    else:
        import pandas as pd
        region_df = pd.DataFrame()
        region_df['geometry'] = [Polygon(vertices[region].tolist()) for region in regions]
        point_df = pd.DataFrame()
        point_df['geometry'] = [Point(pnt) for pnt in points]

    return region_df, point_df

def voronoi_frames(points, radius=None):
    """
    Composite helper to return Voronoi regions and generator points as individual dataframes

    Parameters
    ----------

    points      : array-like
                  originator points


    Returns
    -------

    _           : tuple
                  (region_df, points_df)

                  region_df   : GeoDataFrame (if geopandas available, otherwise Pandas DataFrame)
                                Finite Voronoi polygons as geometries

                  points_df   : GeoDataFrame (if geopandas available, otherwise Pandas DataFrame)
                                Originator points as geometries

    Notes
    -----

    If Geopandas is not available the return types will be Pandas DataFrames
    each with a geometry column populated with PySAL shapes. If Geopandas is
    available, return types are GeoDataFrames with a geometry column populated
    with shapely geometry types.

    Examples
    --------
    >>> points = [(10.2, 5.1), (4.7, 2.2), (5.3, 5.7), (2.7, 5.3)]
    >>> regions_df, points_df = voronoi_frames(points)
    >>> regions_df.shape
    (4, 1)
    >>> regions_df.shape == points_df.shape
    True

    """
    regions, vertices = voronoi(points, radius=radius)
    return as_dataframes(regions, vertices, points)
